# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################

from __future__ import absolute_import
import pytest


@pytest.mark.usefixtures("aiida_profile", "clear_database")
class TestFleurEntrypoints:
    """
    tests all the entry points of the Fleur plugin. Therefore if the plugin is
    recognized by AiiDA and installed right.
    """

    # Calculations

    def test_inpgen_calculation_entry_point(self):
        from aiida.plugins import CalculationFactory
        inpgen_calculation = CalculationFactory('fleur.inpgen')
        assert inpgen_calculation is not None

    def test_fleur_calculation_entry_point(self):
        from aiida.plugins import CalculationFactory

        fleur_calculation = CalculationFactory('fleur.fleur')
        assert fleur_calculation is not None

    # Data

    def test_fleur_fleurinpdata_entry_point(self):
        from aiida.plugins import DataFactory
        from aiida_fleur.data.fleurinp import FleurinpData

        fleurinp = DataFactory('fleur.fleurinp')
        assert fleurinp == FleurinpData

    # Parsers

    def test_inpgen_parser_entry_point(self):
        from aiida.plugins import ParserFactory
        from aiida_fleur.parsers.fleur_inputgen import Fleur_inputgenParser

        parser = ParserFactory('fleur.fleurinpgenparser')
        assert parser == Fleur_inputgenParser

    def test_fleur_parser_entry_point(self):
        from aiida.plugins import ParserFactory
        from aiida_fleur.parsers.fleur import FleurParser

        parser = ParserFactory('fleur.fleurparser')
        assert parser == FleurParser

    # Workflows/workchains

    def test_fleur_scf_wc_entry_point(self):
        from aiida.plugins import WorkflowFactory
        from aiida_fleur.workflows.scf import FleurScfWorkChain

        workflow = WorkflowFactory('fleur.scf')
        assert workflow == FleurScfWorkChain

    def test_fleur_dos_wc_entry_point(self):
        from aiida.plugins import WorkflowFactory
        from aiida_fleur.workflows.dos import fleur_dos_wc

        workflow = WorkflowFactory('fleur.dos')
        assert workflow == fleur_dos_wc

    def test_fleur_band_wc_entry_point(self):
        from aiida.plugins import WorkflowFactory
        from aiida_fleur.workflows.band import fleur_band_wc

        workflow = WorkflowFactory('fleur.band')
        assert workflow == fleur_band_wc

    def test_fleur_eos_wc_entry_point(self):
        from aiida.plugins import WorkflowFactory
        from aiida_fleur.workflows.eos import FleurEosWorkChain

        workflow = WorkflowFactory('fleur.eos')
        assert workflow == FleurEosWorkChain

    # this entry point has currently a problem...
    def test_fleur_initial_cls_wc_entry_point(self):
        from aiida.plugins import WorkflowFactory
        from aiida_fleur.workflows.initial_cls import fleur_initial_cls_wc

        workflow = WorkflowFactory('fleur.init_cls')
        assert workflow == fleur_initial_cls_wc

    def test_fleur_corehole_wc_entry_point(self):
        from aiida.plugins import WorkflowFactory
        from aiida_fleur.workflows.corehole import fleur_corehole_wc

        workflow = WorkflowFactory('fleur.corehole')
        assert workflow == fleur_corehole_wc

    def test_fleur_mae_wc_entry_point(self):
        from aiida.plugins import WorkflowFactory
        from aiida_fleur.workflows.mae import FleurMaeWorkChain

        workflow = WorkflowFactory('fleur.mae')
        assert workflow == FleurMaeWorkChain

    def test_fleur_mae_conv_wc_entry_point(self):
        from aiida.plugins import WorkflowFactory
        from aiida_fleur.workflows.mae_conv import FleurMaeConvWorkChain

        workflow = WorkflowFactory('fleur.mae_conv')
        assert workflow == FleurMaeConvWorkChain

    def test_fleur_ssdisp_wc_entry_point(self):
        from aiida.plugins import WorkflowFactory
        from aiida_fleur.workflows.ssdisp import FleurSSDispWorkChain

        workflow = WorkflowFactory('fleur.ssdisp')
        assert workflow == FleurSSDispWorkChain

    def test_fleur_ssdisp_conv_wc_entry_point(self):
        from aiida.plugins import WorkflowFactory
        from aiida_fleur.workflows.ssdisp_conv import FleurSSDispConvWorkChain

        workflow = WorkflowFactory('fleur.ssdisp_conv')
        assert workflow == FleurSSDispConvWorkChain

    def test_fleur_dmi_wc_entry_point(self):
        from aiida.plugins import WorkflowFactory
        from aiida_fleur.workflows.dmi import FleurDMIWorkChain

        workflow = WorkflowFactory('fleur.dmi')
        assert workflow == FleurDMIWorkChain

    def test_fleur_base_wc_entry_point(self):
        from aiida.plugins import WorkflowFactory
        from aiida_fleur.workflows.base_fleur import FleurBaseWorkChain

        workflow = WorkflowFactory('fleur.base')
        assert workflow == FleurBaseWorkChain

    def test_fleur_base_relax_wc_entry_point(self):
        from aiida.plugins import WorkflowFactory
        from aiida_fleur.workflows.base_relax import FleurBaseRelaxWorkChain

        workflow = WorkflowFactory('fleur.base_relax')
        assert workflow == FleurBaseRelaxWorkChain

    def test_fleur_create_magnetic_wc_entry_point(self):
        from aiida.plugins import WorkflowFactory
        from aiida_fleur.workflows.create_magnetic_film import FleurCreateMagneticWorkChain

        workflow = WorkflowFactory('fleur.create_magnetic')
        assert workflow == FleurCreateMagneticWorkChain



import os
import pytest
import aiida_fleur

aiida_path = os.path.dirname(aiida_fleur.__file__)
TEST_CIF = os.path.join(aiida_path, 'tests/files/cif/AlB.cif')


def test_extract_structure_info(generate_structure,
                                generate_work_chain_node,
                                fixture_localhost):
    """
    I do not test 'extras' here due to some kind of bug
    """
    from aiida_fleur.tools.data_handling import extract_structure_info
    from aiida_fleur.tools.common_aiida import create_group
    from aiida_fleur.tools.read_cif_folder import wf_struc_from_cif
    from aiida.orm import CalcFunctionNode, load_node, CifData
    from aiida.common import LinkType

    pks = []
    uuids = []

    for i in range(3):
        structure_bulk = generate_structure()
        structure_bulk.append_atom(position=(i, 0., -1.99285), symbols='Se')
        structure_bulk.store()
        pks.append(structure_bulk.pk)
        uuids.append(structure_bulk.uuid)

    cif_data = CifData(file=TEST_CIF)
    cif_data.store()
    cif_structure = wf_struc_from_cif(cif_data)

    print(cif_structure.get_incoming().all())

    # create CalcFunction having StructureData input
    calc_function = CalcFunctionNode()
    calc_function.set_attribute('process_label', 'test_label')
    calc_function.add_incoming(structure_bulk, link_type=LinkType.INPUT_CALC,
                               link_label='test_calcfundtion')
    calc_function.store()

    # create WorkChainNode scf having StructureData input
    scf_wc = generate_work_chain_node(computer=fixture_localhost,
                                      entry_point_name='aiida_fleur.scf',
                                      inputs={'structure': load_node(pks[1])})
    scf_wc.set_attribute('process_label', 'fleur_scf_wc')

    # create a group
    group = create_group(name='test_group', nodes=pks[:2],
                         description='test_description')

    result = extract_structure_info(keys=['uuid', 'formula', 'pk', 'symmetry', 'pbc', 'volume',
                                          'total_energy', 'child_nodes', 'natoms', 'group',
                                          'label', 'description', 'cif_file',
                                          'cif_number', 'cif_uuid', 'cif_ref', 'calcfunctions',
                                          'band', 'dos', 'eos', 'init_cls', 'corehole', 'primitive',
                                          'cell', 'scf'])

    # print(result)
    # assert 0
    correct_result = [{
        'formula': 'AlB2',
        'pk': cif_structure.pk,
        'uuid': cif_structure.uuid,
        'natoms': 3,
        'cell': '[[3.009, 0.0, 0.0], [-1.5045, 2.6058704399874, 0.0], [0.0, 0.0, 3.262]]',
        'pbc': '(True, True, True)',
        'label': '',
        'description': '',
        # 'extras': "{{'_aiida_hash': {0}}}".format(cif_structure.get_hash()),
        'symmetry': 'Amm2 (38)',
        'volume': 25.57755127009385,
        'child_nodes': 0,
        'primitive': False,
        'cif_file': ['AlB.cif', cif_data.uuid],
        'group': [],
        'scf': [],
        'band': [],
        'dos': [],
        'eos': [],
        'init_cls': [],
        'corehole': [],
        'calcfunctions': [[], []],
    }, {
        'formula': 'SeSi2',
        'pk': pks[2],
        'uuid': uuids[2],
        'natoms': 3,
        'cell': '[[2.715, 2.715, 0.0], [2.715, 0.0, 2.715], [0.0, 2.715, 2.715]]',
        'pbc': '(True, True, True)',
        'label': '',
        'description': '',
        # 'extras': "{{'_aiida_hash': {0}}}".format(load_node(pks[2]).get_hash()),
        'symmetry': 'P1 (1)',
        'volume': 40.02575174999999,
        'child_nodes': 1,
        'primitive': False,
        'cif_file': ['', ''],
        'group': [],
        'scf': [],
        'band': [],
        'dos': [],
        'eos': [],
        'init_cls': [],
        'corehole': [],
        'calcfunctions': [[calc_function.uuid],
                          ['test_label']],
    }, {
        'formula': 'SeSi2',
        'pk': pks[1],
        'uuid': uuids[1],
        'natoms': 3,
        'cell': '[[2.715, 2.715, 0.0], [2.715, 0.0, 2.715], [0.0, 2.715, 2.715]]',
        'pbc': '(True, True, True)',
        'label': '',
        'description': '',
        # 'extras': "{{'_aiida_hash': {0}}}".format(load_node(pks[1]).get_hash()),
        'symmetry': 'P1 (1)',
        'volume': 40.02575174999999,
        'child_nodes': 1,
        'primitive': False,
        'cif_file': ['', ''],
        'group': ['test_group'],
        'scf': [scf_wc.uuid],
        'band': [],
        'dos': [],
        'eos': [],
        'init_cls': [],
        'corehole': [],
        'calcfunctions': [[], []],
    }, {
        'formula': 'SeSi2',
        'pk': pks[0],
        'uuid': uuids[0],
        'natoms': 3,
        'cell': '[[2.715, 2.715, 0.0], [2.715, 0.0, 2.715], [0.0, 2.715, 2.715]]',
        'pbc': '(True, True, True)',
        'label': '',
        'description': '',
        # 'extras': "{{'_aiida_hash': {0}}}".format(load_node(pks[0]).get_hash),
        'symmetry': 'Imm2 (44)',
        'volume': 40.02575174999999,
        'child_nodes': 0,
        'primitive': False,
        'cif_file': ['', ''],
        'group': ['test_group'],
        'scf': [],
        'band': [],
        'dos': [],
        'eos': [],
        'init_cls': [],
        'corehole': [],
        'calcfunctions': [[], []],
    }]

    for i in result:
        assert i in correct_result



{
    "conv_mode": "density",
    "distance_charge": 0.1406279038,
    "distance_charge_all": [
        61.1110641131,
        43.7556515683,
        ...
    ],
    "distance_charge_units": "me/bohr^3",
    "errors": [],
    "force_diff_last": "can not be determined",
    "force_largest": 0.0,
    "info": [],
    "iterations_total": 23,
    "last_calc_uuid": "b20b5b94-5d80-41a8-82bf-b4d8eee9bddc",
    "loop_count": 1,
    "material": "FePt2",
    "total_energy": -38166.176928494,
    "total_energy_all": [
        -38166.542950054,
        -38166.345602746,
        ...
    ],
    "total_energy_units": "Htr",
    "total_wall_time": 245,
    "total_wall_time_units": "s",
    "warnings": [],
    "workflow_name": "FleurScfWorkChain",
    "workflow_version": "0.4.0"
}



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################

"""
    In this module you find the workflow 'FleurSSDispWorkChain' for the calculation of
    spin spiral dispersion using scalar-relatevistic Hamiltonian.
"""

from __future__ import absolute_import
import copy

import six
from six.moves import map
from lxml import etree

from aiida.engine import WorkChain, ToContext, if_
from aiida.engine import calcfunction as cf
from aiida.orm import Code, load_node, CalcJobNode
from aiida.orm import RemoteData, Dict
from aiida.common import AttributeDict
from aiida.common.exceptions import NotExistent

from aiida_fleur.tools.common_fleur_wf import test_and_get_codenode
from aiida_fleur.tools.common_fleur_wf import get_inputs_fleur
from aiida_fleur.workflows.scf import FleurScfWorkChain
from aiida_fleur.data.fleurinpmodifier import FleurinpModifier
from aiida_fleur.workflows.base_fleur import FleurBaseWorkChain

from aiida_fleur.data.fleurinp import FleurinpData


class FleurSSDispWorkChain(WorkChain):
    """
    This workflow calculates spin spiral dispersion of a structure.
    """

    _workflowversion = "0.2.0"

    _default_options = {
        'resources': {"num_machines": 1, "num_mpiprocs_per_machine": 1},
        'max_wallclock_seconds': 2*60*60,
        'queue_name': '',
        'custom_scheduler_commands': '',
        'import_sys_environment': False,
        'environment_variables': {}
    }

    _wf_default = {
        'beta': {'all': 1.57079},
        'prop_dir': [1.0, 0.0, 0.0],
        'q_vectors': [[0.0, 0.0, 0.0],
                      [0.125, 0.0, 0.0],
                      [0.250, 0.0, 0.0],
                      [0.375, 0.0, 0.0]],
        'ref_qss': [0.0, 0.0, 0.0],
        'serial': False,
        'inpxml_changes': []
    }

    @classmethod
    def define(cls, spec):
        super(FleurSSDispWorkChain, cls).define(spec)
        spec.expose_inputs(FleurScfWorkChain, namespace='scf')
        spec.input("wf_parameters", valid_type=Dict, required=False)
        spec.input("fleur", valid_type=Code, required=True)
        spec.input("remote", valid_type=RemoteData, required=False)
        spec.input("fleurinp", valid_type=FleurinpData, required=False)
        spec.input("options", valid_type=Dict, required=False)

        spec.outline(
            cls.start,
            if_(cls.scf_needed)(
                cls.converge_scf,
                cls.force_after_scf,
            ).else_(
                cls.force_wo_scf,
            ),
            cls.get_results,
            cls.return_results
        )

        spec.output('out', valid_type=Dict)

        # exit codes
        spec.exit_code(230, 'ERROR_INVALID_INPUT_PARAM',
                       message="Invalid workchain parameters.")
        spec.exit_code(231, 'ERROR_INVALID_INPUT_CONFIG',
                       message="Invalid input configuration.")
        spec.exit_code(233, 'ERROR_INVALID_CODE_PROVIDED',
                       message="Invalid code node specified, check inpgen and fleur code nodes.")
        spec.exit_code(235, 'ERROR_CHANGING_FLEURINPUT_FAILED',
                       message="Input file modification failed.")
        spec.exit_code(236, 'ERROR_INVALID_INPUT_FILE',
                       message="Input file was corrupted after user's modifications.")
        spec.exit_code(334, 'ERROR_REFERENCE_CALCULATION_FAILED',
                       message="Reference calculation failed.")
        spec.exit_code(335, 'ERROR_REFERENCE_CALCULATION_NOREMOTE',
                       message="Found no reference calculation remote repository.")
        spec.exit_code(336, 'ERROR_FORCE_THEOREM_FAILED',
                       message="Force theorem calculation failed.")

    def start(self):
        """
        Retrieve and initialize paramters of the WorkChain
        """
        self.report('INFO: started Spin Stiffness calculation workflow version {}\n'
                    ''.format(self._workflowversion))
        self.ctx.info = []
        self.ctx.warnings = []
        self.ctx.errors = []
        self.ctx.energy_dict = []

        # initialize the dictionary using defaults if no wf paramters are given
        wf_default = copy.deepcopy(self._wf_default)
        if 'wf_parameters' in self.inputs:
            wf_dict = self.inputs.wf_parameters.get_dict()
        else:
            wf_dict = wf_default

        extra_keys = []
        for key in wf_dict.keys():
            if key not in wf_default.keys():
                extra_keys.append(key)
        if extra_keys:
            error = 'ERROR: input wf_parameters for SSDisp contains extra keys: {}'.format(
                extra_keys)
            self.report(error)
            return self.exit_codes.ERROR_INVALID_INPUT_PARAM

        # extend wf parameters given by user using defaults
        for key, val in six.iteritems(wf_default):
            wf_dict[key] = wf_dict.get(key, val)
        self.ctx.wf_dict = wf_dict

        if wf_dict['ref_qss'] != wf_dict['q_vectors'][0]:
            error = ("The first q_vector of the forceTheorem step has to be equal to"
                     "the q vector of the reference calculation.")
            self.control_end_wc(error)
            return self.exit_codes.ERROR_INVALID_INPUT_PARAM

        # initialize the dictionary using defaults if no options are given
        defaultoptions = self._default_options
        if 'options' in self.inputs:
            options = self.inputs.options.get_dict()
        else:
            options = defaultoptions

        # extend options given by user using defaults
        for key, val in six.iteritems(defaultoptions):
            options[key] = options.get(key, val)
        self.ctx.options = options

        # Check if user gave valid fleur executable
        inputs = self.inputs
        if 'fleur' in inputs:
            try:
                test_and_get_codenode(inputs.fleur, 'fleur.fleur', use_exceptions=True)
            except ValueError:
                error = "The code you provided for FLEUR does not use the plugin fleur.fleur"
                self.control_end_wc(error)
                return self.exit_codes.ERROR_INVALID_CODE_PROVIDED

        # Check if user gave an input setup making any sense
        if inputs.scf:
            self.ctx.scf_needed = True
            if 'remote' in inputs:
                error = "ERROR: you gave SCF input + remote for the FT"
                self.control_end_wc(error)
                return self.exit_codes.ERROR_INVALID_INPUT_CONFIG
            if 'fleurinp' in inputs:
                error = "ERROR: you gave SCF input + fleurinp for the FT"
                self.control_end_wc(error)
                return self.exit_codes.ERROR_INVALID_INPUT_CONFIG
        elif 'remote' not in inputs:
            error = "ERROR: you gave neither SCF input nor remote for the FT"
            self.control_end_wc(error)
            return self.exit_codes.ERROR_INVALID_INPUT_CONFIG
        else:
            self.ctx.scf_needed = False

    def scf_needed(self):
        """
        Returns True if SCF WC is needed.
        """
        return self.ctx.scf_needed

    def converge_scf(self):
        """
        Converge charge density for collinear case which is a reference for futher
        spin spiral calculations.
        """
        inputs = self.get_inputs_scf()
        res = self.submit(FleurScfWorkChain, **inputs)
        return ToContext(reference=res)

    def get_inputs_scf(self):
        """
        Initialize inputs for the scf cycle
        """
        input_scf = AttributeDict(self.exposed_inputs(FleurScfWorkChain, namespace='scf'))

        if 'wf_parameters' not in input_scf:
            scf_wf_dict = {}
        else:
            scf_wf_dict = input_scf.wf_parameters.get_dict()

        if 'inpxml_changes' not in scf_wf_dict:
            scf_wf_dict['inpxml_changes'] = []
        # set up q vector for the reference calculation
        list_ref_qss = self.ctx.wf_dict['ref_qss']
        if [x for x in list_ref_qss if x != 0]:
            changes_dict = {'qss': self.ctx.wf_dict['ref_qss'],
                            'l_noco': True,
                            'ctail': False,
                            'l_ss': True}
        else:
            changes_dict = {'qss': ' 0.0 0.0 0.0 ',
                            'l_noco': False,
                            'ctail': True,
                            'l_ss': False}

        scf_wf_dict['inpxml_changes'].append(('set_inpchanges', {'change_dict': changes_dict}))

        # change beta parameter
        for key, val in six.iteritems(self.ctx.wf_dict.get('beta')):
            scf_wf_dict['inpxml_changes'].append(
                ('set_atomgr_att_label',
                 {'attributedict': {'nocoParams': [('beta', val)]},
                  'atom_label': key
                  }))

        input_scf.wf_parameters = Dict(dict=scf_wf_dict)

        if 'structure' in input_scf:  # add info about spin spiral propagation
            if 'calc_parameters' in input_scf:
                calc_parameters = input_scf.calc_parameters.get_dict()
            else:
                calc_parameters = {}
            calc_parameters['qss'] = {'x': self.ctx.wf_dict['prop_dir'][0],
                                      'y': self.ctx.wf_dict['prop_dir'][1],
                                      'z': self.ctx.wf_dict['prop_dir'][2]}
            input_scf.calc_parameters = Dict(dict=calc_parameters)

        return input_scf

    def change_fleurinp(self):
        """
        This routine sets somethings in the fleurinp file before running a fleur
        calculation.
        """
        self.report('INFO: run change_fleurinp')

        if self.ctx.scf_needed:  # use fleurinp from scf wc
            try:
                fleurin = self.ctx.reference.outputs.fleurinp
            except NotExistent:
                error = 'Fleurinp generated in the reference calculation is not found.'
                self.control_end_wc(error)
                return self.exit_codes.ERROR_REFERENCE_CALCULATION_FAILED
        else:
            if 'fleurinp' in self.inputs:  # use the given fleurinp
                fleurin = self.inputs.fleurinp
            else:  # or generate a new one from inp.xml located in the remote folder
                remote_node = self.inputs.remote
                for link in remote_node.get_incoming().all():
                    if isinstance(link.node, CalcJobNode):
                        parent_calc_node = link.node
                retrieved_node = parent_calc_node.get_outgoing().get_node_by_label('retrieved')
                fleurin = FleurinpData(files=['inp.xml'], node=retrieved_node)

        # copy inpchanges from wf parameters
        fchanges = self.ctx.wf_dict.get('inpxml_changes', [])
        # create forceTheorem tags
        fchanges.extend([('create_tag',
                          {'xpath': '/fleurInput',
                           'newelement': 'forceTheorem'
                           }),
                         ('create_tag',
                          {'xpath': '/fleurInput/forceTheorem',
                           'newelement': 'spinSpiralDispersion'
                           })])

        for i, vectors in enumerate(self.ctx.wf_dict['q_vectors']):

            fchanges.append(('create_tag',
                             {'xpath': '/fleurInput/forceTheorem/spinSpiralDispersion',
                              'newelement': 'q'
                              }))
            fchanges.append(('xml_set_text_occ',
                             {'xpathn': '/fleurInput/forceTheorem/spinSpiralDispersion/q',
                              'text': ' '.join(map(str, vectors)),
                              'create': False,
                              'occ': i
                              }))

        changes_dict = {'itmax': 1,
                        'l_noco': True,
                        'ctail': False,
                        'l_ss': True}
        fchanges.append(('set_inpchanges', {'change_dict': changes_dict}))

        # change beta parameter
        for key, val in six.iteritems(self.ctx.wf_dict.get('beta')):
            fchanges.append(('set_atomgr_att_label',
                             {'attributedict': {'nocoParams': [('beta', val)]},
                              'atom_label': key
                              }))

        if fchanges:  # change inp.xml file
            fleurmode = FleurinpModifier(fleurin)
            avail_ac_dict = fleurmode.get_avail_actions()

            # apply further user dependend changes
            for change in fchanges:
                function = change[0]
                para = change[1]
                method = avail_ac_dict.get(function, None)
                if not method:
                    error = ("ERROR: Input 'inpxml_changes', function {} "
                             "is not known to fleurinpmodifier class, "
                             "please check/test your input. I abort..."
                             "".format(function))
                    self.control_end_wc(error)
                    return self.exit_codes.ERROR_CHANGING_FLEURINPUT_FAILED

                else:  # apply change
                    method(**para)

            # validate
            try:
                fleurmode.show(display=False, validate=True)
            except etree.DocumentInvalid:
                error = ('ERROR: input, user wanted inp.xml changes did not validate')
                self.control_end_wc(error)
                return self.exit_codes.ERROR_INVALID_INPUT_FILE

            # apply
            out = fleurmode.freeze()
            self.ctx.fleurinp = out
            return
        else:  # otherwise do not change the inp.xml
            self.ctx.fleurinp = fleurin
            return

    def force_after_scf(self):
        '''
        This routine uses the force theorem to calculate energies dispersion of
        spin spirals. The force theorem calculations implemented into the FLEUR
        code. Hence a single iteration FLEUR input file having <forceTheorem> tag
        has to be created and submitted.
        '''
        calc = self.ctx.reference

        if not calc.is_finished_ok:
            message = ('The reference SCF calculation was not successful.')
            self.control_end_wc(message)
            return self.exit_codes.ERROR_REFERENCE_CALCULATION_FAILED

        try:
            outpara_node = calc.outputs.output_scf_wc_para
        except NotExistent:
            message = ('The reference SCF calculation failed, no scf output node.')
            self.control_end_wc(message)
            return self.exit_codes.ERROR_REFERENCE_CALCULATION_FAILED

        outpara = outpara_node.get_dict()

        if 'total_energy' not in outpara:
            message = ('Did not manage to extract float total energy from the'
                       ' reference SCF calculation.')
            self.control_end_wc(message)
            return self.exit_codes.ERROR_REFERENCE_CALCULATION_FAILED

        self.report('INFO: run Force theorem calculations')

        status = self.change_fleurinp()
        if status:
            return status

        fleurin = self.ctx.fleurinp

        # Do not copy mixing_history* files from the parent
        settings = {'remove_from_remotecopy_list': ['mixing_history*']}

        # Retrieve remote folder of the reference calculation
        pk_last = 0
        scf_ref_node = load_node(calc.pk)
        for i in scf_ref_node.called:
            if i.node_type == 'process.workflow.workchain.WorkChainNode.':
                if i.process_class is FleurBaseWorkChain:
                    if pk_last < i.pk:
                        pk_last = i.pk
        try:
            remote = load_node(pk_last).outputs.remote_folder
        except AttributeError:
            message = ('Found no remote folder of the reference scf calculation.')
            self.control_end_wc(message)
            return self.exit_codes.ERROR_REFERENCE_CALCULATION_NOREMOTE

        label = 'Force_theorem_calculation'
        description = 'This is a force theorem calculation for all SQA'

        code = self.inputs.fleur
        options = self.ctx.options.copy()

        inputs_builder = get_inputs_fleur(code, remote,
                                          fleurin, options, label, description, settings,
                                          serial=self.ctx.wf_dict['serial'])
        future = self.submit(FleurBaseWorkChain, **inputs_builder)
        return ToContext(f_t=future)

    def force_wo_scf(self):
        """
        Submit FLEUR force theorem calculation using input remote
        """
        self.report('INFO: run Force theorem calculations')

        status = self.change_fleurinp()
        if status:
            return status

        fleurin = self.ctx.fleurinp

        # Do not copy mixing_history* files from the parent
        settings = {'remove_from_remotecopy_list': ['mixing_history*']}

        # Retrieve remote folder from the inputs
        remote = self.inputs.remote

        label = 'Force_theorem_calculation'
        description = 'This is a force theorem calculation for all SQA'

        code = self.inputs.fleur
        options = self.ctx.options.copy()

        inputs_builder = get_inputs_fleur(code, remote,
                                          fleurin, options, label, description, settings,
                                          serial=self.ctx.wf_dict['serial'])
        future = self.submit(FleurBaseWorkChain, **inputs_builder)
        return ToContext(f_t=future)

    def get_results(self):
        """
        Generates results of the workchain.
        """
        t_energydict = []
        htr_to_ev = 27.21138602
        try:
            calculation = self.ctx.f_t
            if not calculation.is_finished_ok:
                message = ('ERROR: Force theorem Fleur calculation failed somehow it has '
                           'exit status {}'.format(calculation.exit_status))
                self.control_end_wc(message)
                return self.exit_codes.ERROR_FORCE_THEOREM_FAILED
        except AttributeError:
            message = 'ERROR: Something went wrong I do not have a force theorem Fleur calculation'
            self.control_end_wc(message)
            return self.exit_codes.ERROR_FORCE_THEOREM_FAILED

        try:
            out_dict = calculation.outputs.output_parameters.dict
            t_energydict = out_dict.spst_force_evSum
            e_u = out_dict.energy_units

            # Find a minimal value of SpSp and count it as 0
            minenergy = min(t_energydict)

            if e_u == 'Htr' or 'htr':
                t_energydict = [htr_to_ev * (x-minenergy) for x in t_energydict]
            else:
                t_energydict = [(x-minenergy) for x in t_energydict]

            self.ctx.energy_dict = t_energydict

        except AttributeError:
            message = ('Did not manage to read evSum or energy units after FT calculation.')
            self.control_end_wc(message)
            return self.exit_codes.ERROR_FORCE_THEOREM_FAILED

    def return_results(self):
        """
        This function outputs results of the wc
        """
        out = {'workflow_name': self.__class__.__name__,
               'workflow_version': self._workflowversion,
               # 'initial_structure': self.inputs.structure.uuid, # parse from fleurinp or remove
               'is_it_force_theorem': True,
               'energies': self.ctx.energy_dict,
               'q_vectors': self.ctx.wf_dict['q_vectors'],
               'energy_units': 'eV',
               'info': self.ctx.info,
               'warnings': self.ctx.warnings,
               'errors': self.ctx.errors
               }

        out = save_output_node(Dict(dict=out))
        self.out('out', out)

    def control_end_wc(self, errormsg):
        """
        Controlled way to shutdown the workchain. It will initialize the output nodes
        The shutdown of the workchain will has to be done afterwards
        """
        self.report(errormsg)  # because return_results still fails sometimes
        self.ctx.errors.append(errormsg)
        self.return_results()


@cf
def save_output_node(out):
    """
    This calcfunction saves the out dict in the db
    """
    out_wc = out.clone()
    return out_wc





# get_econfig
from __future__ import absolute_import
def test_get_econfig_W():
    from aiida_fleur.tools.element_econfig_list import get_econfig

    assert get_econfig('W', full=False) == '[Kr] 5s2 4d10 4f14 | 5p6 6s2 5d4'
    assert get_econfig('W', full=True) == '1s2 2s2 2p6 3s2 3p6 3d10 4s2 4p6 5s2 4d10 4f14 | 5p6 6s2 5d4'

# get_coreconfig
def test_get_coreconfig_W():
    from aiida_fleur.tools.element_econfig_list import get_coreconfig
    
    assert get_coreconfig('W', full=False) == '[Kr] 5s2 4d10 4f14'
    assert get_coreconfig('W', full=True) == '1s2 2s2 2p6 3s2 3p6 3d10 4s2 4p6 5s2 4d10 4f14'

# rek_econ
def test_rek_econ_interface_W():
    from aiida_fleur.tools.element_econfig_list import rek_econ

    assert rek_econ('[Kr] 4d10 4f14 5p6 | 5s2 6s2 5d4') == '1s2 2s2 2p6 3s2 3p6 3d10 4s2 4p6 4d10 4f14 5p6 | 5s2 6s2 5d4'

# convert_fleur_config_to_econfig
def test_convert_fleur_econfig_to_econfig_interface():
    from aiida_fleur.tools.element_econfig_list import convert_fleur_config_to_econfig

    config1 = '[Kr] (4d3/2) (4d5/2) (4f5/2) (4f7/2)'
    config2 = '[Kr]'
    config3 = '[Kr] 4d3/2 4d5/2 4f5/2'
    
    res1 = '[Kr] 4d10 4f14'
    res1_1 = '[Kr] 4d3/2 4d5/2 4f5/2 4f7/2'
    res2 = '[Kr]'
    res3 = '[Kr] 4d10 4f6'
    res3_1 = '[Kr] 4d3/2 4d5/2 4f5/2'

    assert convert_fleur_config_to_econfig(config1) == res1
    assert convert_fleur_config_to_econfig(config1, keep_spin=True) == res1_1
    assert convert_fleur_config_to_econfig(config2) == res2
    assert convert_fleur_config_to_econfig(config3) == res3
    assert convert_fleur_config_to_econfig(config3, keep_spin=True) == res3_1

# highest_unocc_valence
def test_rek_econ_interface_W():
    from aiida_fleur.tools.element_econfig_list import highest_unocc_valence
    assert highest_unocc_valence('[Kr] 4d10 4f14 5p6 | 5s2 6s2 5d4') == '5d4'
    assert highest_unocc_valence('1s2 | 2s2') == '2p0'  


# get_spin_econfig
def test_get_spin_econfig_interface():
    from aiida_fleur.tools.element_econfig_list import get_spin_econfig
    
    # basic behavior
    assert get_spin_econfig('1s2 2s2 2p6') == '1s1/2 2s1/2 2p1/2 2p3/2'
    
    # does autocompletion work
    res1 = '1s1/2 2s1/2 2p1/2 2p3/2 3s1/2 3p1/2 3p3/2 3d3/2 3d5/2 4s1/2 4p1/2 4p3/2 4d3/2 4d5/2 4f5/2 4f7/2'
    assert get_spin_econfig('[Kr] 4d10 4f14') == res1

    # if addition in spin config given, does this work? and no doubles?
    res2 = '1s1/2 2s1/2 2p1/2 2p3/2 3s1/2 3p1/2 3p3/2 3d3/2 3d5/2 4s1/2 4p1/2 4p3/2 4d3/2 4d5/2 4f5/2 4f7/2'
    assert get_spin_econfig('[Kr] 4d3/2 4d5/2 4f5/2 4f7/2') == res2

# econfig_str_hole
def test_econfig_str_hole_1s_Be():
    from aiida_fleur.tools.element_econfig_list import econfigstr_hole
    
    assert econfigstr_hole('1s2 | 2s2', '1s2', '2p0') == '1s1 | 2s2 2p1'


# get_state_occ

def test_get_state_occ_interface_half_valence_hole_W():
    from aiida_fleur.tools.element_econfig_list import get_state_occ
    econfig = '1s2 2s2 2p6 3s2 3p6 3d10 4s2 4p6 4d10 4f13 5p6 | 5s2 6s2 5d5'
    assert get_state_occ(econfig, corehole='4f 5/2', valence='5d', ch_occ=0.5)[1] == {'state' : '(4f5/2)', 'spinUp' : 3.0, 'spinDown' : 2.5}
    assert get_state_occ(econfig, corehole='4f 5/2', valence='5d', ch_occ=0.5)[0] == {'state' : '(5d5/2)', 'spinUp' : 0.5, 'spinDown' : 0.0}



# -*- coding: utf-8 -*-
#
# Fleur plugin documentation build configuration file, created by
# sphinx-quickstart on Wed Dec  7 16:39:12 2016.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import sys
import os
import time
# Following 3 lines avoid the need of importing load_dbenv() for compiling the
# documentation -> works also without verdi install
sys.path.append( os.path.join( os.path.split(__file__)[0],
                                           os.pardir,os.pardir) )
sys.path.append( os.path.join( os.path.split(__file__)[0],
                                           os.pardir))
os.environ['DJANGO_SETTINGS_MODULE'] = 'rtd_settings'

import aiida
import aiida_fleur


# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#sys.path.insert(0, os.path.abspath('../aiida_fleur'))

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
needs_sphinx = '1.5'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ['sphinx.ext.autodoc',
              'sphinx.ext.mathjax',
              'sphinx.ext.doctest', 
              'sphinx.ext.todo', 
              'sphinx.ext.coverage',  
              'sphinx.ext.ifconfig', 
              'sphinx.ext.intersphinx',
              'sphinx.ext.viewcode']

todo_include_todos = True

intersphinx_mapping = {
#    'python': ('https://docs.python.org/2.7', None),
     'aiida': ('http://aiida_core.readthedocs.io/en/latest/', None),
    }

nitpick_ignore = [('py:obj', 'module')]


# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The encoding of source files.
#source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u'AiiDA-FLEUR'
#copyright = u'2016-2019, Jens Broeder'
copyright_first_year = 2016
copyright_owners = u"FZ Jülich GmbH, Germany"

current_year = time.localtime().tm_year
copyright_year_string = current_year if current_year == copyright_first_year else "{}-{}".format(copyright_first_year, current_year)
copyright = u'{}, {}. All rights reserved'.format(copyright_year_string, copyright_owners)


# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
# The full version, including alpha/beta/rc tags.

release = aiida_fleur.__version__
# The short X.Y version.
version = '.'.join(release.split('.')[:2])

author = "The AiiDA-FLEUR team."

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
#today = ''
# Else, today_fmt is used as the format for a strftime call.
#today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build']

# The reST default role (used for this markup: `text`) to use for all
# documents.
#default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
#add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
#add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
#show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
#modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
#keep_warnings = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'sphinx_rtd_theme' #'default'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#html_theme_options = {}

# Add any paths that contain custom themes here, relative to this directory.
#html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
#html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
#html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
#html_logo = None

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
#html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
#html_extra_path = []

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
#html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
#html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
#html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
#html_additional_pages = {}

# If false, no module index is generated.
#html_domain_indices = True

# If false, no index is generated.
#html_use_index = True

# If true, the index is split into individual pages for each letter.
#html_split_index = False

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = False

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
#html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
#html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
html_use_opensearch = 'http://aiida-fleur.readthedocs.io'

# This is the file name suffix for HTML files (e.g. ".xhtml").
#html_file_suffix = None

# Language to be used for generating the HTML full-text search index.
# Sphinx supports the following languages:
#   'da', 'de', 'en', 'es', 'fi', 'fr', 'hu', 'it', 'ja'
#   'nl', 'no', 'pt', 'ro', 'ru', 'sv', 'tr'
html_search_language = 'en'

# Output file base name for HTML help builder.
htmlhelp_basename = 'aiida-fleurdoc'


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
# The paper size ('letterpaper' or 'a4paper').
#'papersize': 'letterpaper',

# The font size ('10pt', '11pt' or '12pt').
#'pointsize': '10pt',

# Additional stuff for the LaTeX preamble.
#'preamble': '',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
  ('index', 'aiida-fleur.tex', u'AiiDa-FLEUR Documentation',
   author.replace(',',r'\and'), 'manual'),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
#latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
#latex_use_parts = False

# If true, show page references after internal links.
#latex_show_pagerefs = False

# If true, show URL addresses after external links.
#latex_show_urls = False

# Documents to append as an appendix to all manuals.
#latex_appendices = []

# If false, no module index is generated.
#latex_domain_indices = True


# on_rtd is whether we are on readthedocs.org, this line of code grabbed
# from docs.readthedocs.org
# NOTE: it is needed to have these lines before load_dbenv()
on_rtd = os.environ.get('READTHEDOCS', None) == 'True'

# Following 3 lines avoid the need of importing load_dbenv() for compiling the
# documentation -> works also without verdi install
#sys.path.append( os.path.join( os.path.split(__file__)[0],
#                                           os.pardir,os.pardir) )
#sys.path.append( os.path.join( os.path.split(__file__)[0],
#                                           os.pardir))
#os.environ['DJANGO_SETTINGS_MODULE'] = 'rtd_settings'

if not on_rtd:  # only import and set the theme if we're building docs locally
    try:
        import sphinx_rtd_theme
        html_theme = 'sphinx_rtd_theme'
        html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]
    except ImportError:
        # No sphinx_rtd_theme installed
        pass
    # Load the database environment by first loading the profile and then loading the backend through the manager
    from aiida.manage.configuration import get_config, load_profile
    from aiida.manage.manager import get_manager
    config = get_config()
    load_profile(config.default_profile_name)
    get_manager().get_backend()
else:
    # Back-end settings for readthedocs online documentation.
    from aiida.manage import configuration
    configuration.IN_RT_DOC_MODE = True
    configuration.BACKEND = "django"
    configuration.AIIDADB_PROFILE = "default"


'''
def run_apidoc(_):
    """Runs sphinx-apidoc when building the documentation.

    Needs to be done in conf.py in order to include the APIdoc in the
    build on readthedocs.

    See also https://github.com/rtfd/readthedocs.org/issues/1139
    """
    source_dir = os.path.abspath(os.path.dirname(__file__))
    apidoc_dir = os.path.join(source_dir, 'apidoc')
    package_dir = os.path.join(source_dir, os.pardir, os.pardir, 'aiida_fleur')

    # In #1139, they suggest the route below, but for me this ended up
    # calling sphinx-build, not sphinx-apidoc
    #from sphinx.apidoc import main
    #main([None, '-e', '-o', apidoc_dir, package_dir, '--force'])

    import subprocess
    cmd_path = 'sphinx-apidoc'
    if hasattr(sys, 'real_prefix'):  # Check to see if we are in a virtualenv
        # If we are, assemble the path manually
        cmd_path = os.path.abspath(os.path.join(sys.prefix, 'bin', 'sphinx-apidoc'))

    options = [
        '-o', apidoc_dir, package_dir,
        '--private',
        '--force',
        '--no-headings',
        '--module-first',
        '--no-toc',
        '--maxdepth', '4',
    ]

    # See https://stackoverflow.com/a/30144019
    env = os.environ.copy()
    env["SPHINX_APIDOC_OPTIONS"] = 'members,special-members,private-members,undoc-members,show-inheritance'
    subprocess.check_call([cmd_path] + options, env=env)

def setup(app):
    app.connect('builder-inited', run_apidoc)
'''

# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ('index', 'aiida-fleur', u'AiiDA-FLEUR Documentation',
     [author], 1)
]

# If true, show URL addresses after external links.
#man_show_urls = False


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
  ('index', 'aiida-fleur', u'AiiDA-Fleur Documentation',
   author, 'aiida-fleur', 'One line description of project.',
   'Miscellaneous'),
]

# Documents to append as an appendix to all manuals.
#texinfo_appendices = []

# If false, no module index is generated.
#texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
#texinfo_show_urls = 'footnote'

# If true, do not generate a @detailmenu in the "Top" node's menu.
#texinfo_no_detailmenu = False

# -- Options for Epub output ---------------------------------------------------

# Bibliographic Dublin Core info.
epub_title = u'AiiDA-FLEUR'
epub_author = author
epub_publisher = author
epub_copyright = copyright

# The language of the text. It defaults to the language option
# or en if the language is not set.
#epub_language = ''

# The scheme of the identifier. Typical schemes are ISBN or URL.
#epub_scheme = ''

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#epub_identifier = ''

# A unique identification for the text.
#epub_uid = ''

# A tuple containing the cover image and cover page html template filenames.
#epub_cover = ()

# HTML files that should be inserted before the pages created by sphinx.
# The format is a list of tuples containing the path and title.
#epub_pre_files = []

# HTML files shat should be inserted after the pages created by sphinx.
# The format is a list of tuples containing the path and title.
#epub_post_files = []

# A list of files that should not be packed into the epub file.
#epub_exclude_files = []

# The depth of the table of contents in toc.ncx.
#epub_tocdepth = 3


# Warnings to ignore when using the -n (nitpicky) option
# We should ignore any python built-in exception, for instance
nitpick_ignore = []

for line in open('nitpick-exceptions'):
    if line.strip() == "" or line.startswith("#"):
        continue
    dtype, target = line.split(None, 1)
    target = target.strip()
    nitpick_ignore.append((dtype, target))



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################

"""
In this module you find the workflow 'fleur_optimize_parameter_wc', which finds
working/(in the future, optiomal) flapw parameters for a given Structure
"""

#import numpy as np
from __future__ import absolute_import
from aiida.plugins import DataFactory
from aiida.orm import Code, load_node
from aiida.orm import StructureData, RemoteData, Dict
#from aiida.orm.data.base import Float
#from aiida.work.process_registry import ProcessRegistry
from aiida.engine import WorkChain, ToContext#,Outputs
#from aiida.work import calcfunction as cf
from aiida.engine import submit
#from aiida_fleur.tools.StructureData_util import rescale, is_structure
#from aiida_fleur.workflows.scf import fleur_scf_wc
from aiida_fleur.calculation.fleurinputgen import FleurinputgenCalculation
#from aiida_fleur.data.fleurinpmodifier import FleurinpModifier
from aiida_fleur.tools.common_fleur_wf import get_inputs_fleur, get_inputs_inpgen
from aiida_fleur.tools.common_fleur_wf import test_and_get_codenode
import six

from aiida_fleur.data.fleurinp import FleurinpData


class fleur_optimize_parameters_wc(WorkChain):
    """
    This workflow finds out working/(in the future optimal) flapw parameters
    from a structure. For now, it runs inpgen on the structure and uses
    the Fleur defaults.

    :param wf_parameters: Dict node, optional, protocol specification
                          will be parsed like this to fleur_eos_wc
    :param structure: StructureData node, bulk crystal structure
    :param inpgen: Code node,
    :param fleur: Code node,

    :return output_optimized_wc_para: Dict node, contains relevant output information
                                      about general succces.
    :return optimized_para: Dict node usable by inpgen
    :return optimized_fleurinp: FleurinpData with optimized

    """

    _workflowversion = "0.1.0"

    def __init__(self, *args, **kwargs):
        super(fleur_optimize_parameters_wc, self).__init__(*args, **kwargs)

    @classmethod
    def define(cls, spec):
        super(fleur_optimize_parameters_wc, cls).define(spec)
        spec.input("wf_parameters", valid_type=Dict, required=False,
                   default=Dict(dict={
                       'resources' : {"num_machines": 1},#, "num_mpiprocs_per_machine" : 12},
                       'walltime_sec':  60*60,
                       'queue_name' : '',
                       'custom_scheduler_commands' : ''}))
        spec.input("structure", valid_type=StructureData, required=True)
        spec.input("inpgen", valid_type=Code, required=True)
        spec.input("fleur", valid_type=Code, required=False)
        spec.outline(
            cls.start,
            cls.run_inpgen,
            cls.determine_parameters,
            cls.return_results
        )
        #spec.dynamic_output()


    def start(self):
        """
        check parameters, what condictions? complete?
        check input nodes
        """
        self.report('started fleur_optimize_parameter workflow version {}'.format(self._workflowversion))
        self.report("Workchain node identifiers: ")#{}".format(ProcessRegistry().current_calc_node))

        ### input check ###

        # initialize contexts

        self.ctx.successful = True
        # Check on inputnodes

        inputs = self.inputs

        # wf_parameters:
        wf_dict = inputs.wf_parameters.get_dict()

        # set values, or DEFAULTS
        self.ctx.serial = wf_dict.get('serial', False)
        self.ctx.custom_scheduler_commands = wf_dict.get('custom_scheduler_commands', '')
        self.ctx.max_number_runs = wf_dict.get('fleur_runmax', 4)
        self.ctx.description_wf = self.inputs.get('_description', '') + '|fleur_optimize_parameters_wc|'
        self.ctx.walltime_sec = wf_dict.get('walltime_sec', 60*60)
        self.ctx.resources = wf_dict.get('resources' , {"num_machines": 1})
        self.ctx.queue = wf_dict.get('queue_name' , '')

        # codes
        if 'inpgen' in inputs:
            try:
                test_and_get_codenode(inputs.inpgen, 'fleur.inpgen',
                                      use_exceptions=True)
            except ValueError:
                error = ("The code you provided for inpgen of FLEUR does not "
                         "use the plugin fleur.inpgen")
                self.control_end_wc(error)
                self.abort(error)

        if 'fleur' in inputs:
            try:
                test_and_get_codenode(inputs.fleur, 'fleur.fleur',
                                      use_exceptions=True)
            except ValueError:
                error = ("The code you provided for FLEUR does not "
                         "use the plugin fleur.fleur")
                self.control_end_wc(error)
                self.abort(error)


    def run_inpgen(self):
        """
        So far run inpgen and see what you get
        """

        structure = self.inputs.structure
        self.ctx.formula = structure.get_formula()
        label = 'scf: inpgen'
        description = '{} inpgen on {}'.format(self.ctx.description_wf,
                                               self.ctx.formula)

        inpgencode = self.inputs.inpgen
        if 'calc_parameters' in self.inputs:
            params = self.inputs.calc_parameters
        else:
            params = None

        options = {"max_wallclock_seconds": self.ctx.walltime_sec,
                   "resources": self.ctx.resources,
                   "queue_name" : self.ctx.queue}

        inputs = get_inputs_inpgen(structure, inpgencode, options, label,
                                   description, params=params)
        self.report('INFO: run inpgen')
        future = submit(FleurinputgenCalculation, **inputs)

        return ToContext(inpgen=future, last_calc=future)


    def determine_parameters(self):
        """
        Determine_parameters, so far extract them from FleurinpData
        """

        try:
            fleurin = self.ctx['inpgen'].outputs.fleurinpData
        except AttributeError:
            error = 'No fleurinpData found, inpgen failed'
            self.control_end_wc(error)

        # TODO this has to be implemented
        self.ctx.optimal_parameters = fleurin.get_parameterdata(fleurin)


    # TODO determine optimasation
    # look at number of atoms
    # basis size nvd not known, only after fleur run
    # can this be examined via kmax?
    # number of kpoints
    #

    def return_results(self):
        """
        Prepare and return the result nodes
        """

        if self.ctx.optimal_parameters:
            optimal_parameters_uuid = self.ctx.optimal_parameters.uuid
        else:
            optimal_parameters_uuid = None

        out = {'workflow_name' : self.__class__.__name__,
               'workflow_version' : self._workflowversion,
               'structure': self.inputs.structure.uuid,
               'optimal_para' : optimal_parameters_uuid,
               'successful' : self.ctx.successful}

        outnode = Dict(dict=out)

        returndict = {}

        if self.ctx.successful:
            self.report('Done, fleur_optimize_parameter_wc calculation complete')

            returndict['output_optimized_wc_para'] = outnode
            returndict['optimized_para'] = self.ctx.optimal_parameters
            #optimized_fleurinp
        else:
            self.report('Done, but something failed in fleur_optimize_parameter_wc.')


        # create link to workchain node
        for link_name, node in six.iteritems(returndict):
            self.out(link_name, node)

    def control_end_wc(self, errormsg):
        """
        Controled way to shutdown the workchain. Can initalize the output nodes,
        or set things in the contents
        """
        self.ctx.successful = False
        self.ctx.abort = True
        self.report(errormsg) # because return_results still fails somewhen
        self.return_results()
        #self.abort_nowait(errormsg)
        self.abort(errormsg)



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################

"""
In this module you find the worklfow 'fleur_delta_wc' which is a turnkey solution to calculate a delta for a given code with AiiDA.
"""
#TODO: calculation of delta value from the files
# submit everything if subworkchaining works in Aiida
# parameter node finding is not optimal.

# TODO several eos starts wich only 20 structures to limit jobs throughput
from __future__ import absolute_import
from __future__ import print_function
import os
from string import digits
#from pprint import pprint

from aiida.plugins import DataFactory
from aiida.orm import Code, Group
from aiida.orm import RemoteData, StructureData, Dict, SinglefileData
from aiida.engine import WorkChain, ToContext#, while_
#from aiida.work.process_registry import ProcessRegistry
from aiida.engine import calcfunction as cf
from aiida.engine import submit
from aiida.common.exceptions import NotExistent
from aiida_fleur.workflows.eos import FleurEosWorkChain
import six

#from aiida_fleur.tools.xml_util import eval_xpath2
#from lxml import etree

from aiida_fleur.data.fleurinp import FleurinpData


class fleur_delta_wc(WorkChain):
    """
    This workflow calculates a equation of states and from a given
    group of structures in the database using a group of given parameter nodes in the database
    """

    _workflowversion = "0.3.2"
    _wf_default = {}


    @classmethod
    def define(cls, spec):
        super(fleur_delta_wc, cls).define(spec)
        spec.input("wf_parameters", valid_type=Dict, required=False,
                   default=Dict(dict={'struc_group': 'delta',
                                               'para_group' : 'delta',
                                               'add_extra' : {'type' : 'delta run'},
                                               #'group_label' : 'delta_eos',
                                               'joblimit' : 100,
                                               'part' : [1,2,3,4],
                                               'points' : 7,
                                               'step' : 0.02}))
        spec.input("options", valid_type=Dict, required=False,
                   default=Dict(dict={
                            'resources': {"num_machines": 1},
                            'walltime_sec': 60*60,
                            'queue_name': '',
                            'custom_scheduler_commands' : '',
                            'import_sys_environment' : False,
                            'environment_variables' : {}}))
        spec.input("inpgen", valid_type=Code, required=True)
        spec.input("fleur", valid_type=Code, required=True)
        spec.outline(
            cls.start_up,
            #while_(cls.calculations_left_torun)(
            cls.run_eos,#),
            cls.extract_results_eos,
            cls.calculate_delta,
            cls.return_results,

        )
        #spec.dynamic_output()

    def start_up(self):
        """
        init context and some parameters
        """

        #print('started delta workflow version {}'.format(self._workflowversion))
        #print("Workchain node identifiers: {}".format(ProcessRegistry().current_calc_node))
        #identifier =  0#ProcessRegistry().current_calc_node
        #self.ctx.own_uuid = identifier.uuid
        #self.ctx.own_pk = identifier.pk

        self.report('started delta workflow version {} with identifier: '#{}'
                    ''.format(self._workflowversion))#, identifier))

        # init
        self.ctx.calcs_to_run = []
        # input  check

        # check if right codes
        wf_dict = self.inputs.wf_parameters.get_dict()
        options_dict = self.inputs.get('options', Dict(dict={'resources' : {"num_machines": 1}, 'walltime_sec': int(5.5*3600)}))
        self.ctx.inputs_eos = {
            'fleur': self.inputs.fleur,
            'inpgen': self.inputs.inpgen,
            'wf_parameters':
                {'points' : wf_dict.get('points', 7),
                 'step' : wf_dict.get('step', 0.02),
                 'guess' : 1.0},
            'options' : options_dict
             }
        self.ctx.wc_eos_para = Dict(dict=self.ctx.inputs_eos.get('wf_parameters'))
        self.ctx.ncalc = 1 # init
        self.get_calcs_from_groups()
        self.ctx.successful = True
        self.ctx.warnings = []
        self.ctx.labels = []
        #self.ctx.calcs_to_run = calcs
        self.ctx.ncalcs = len(self.ctx.calcs_to_run)
        print((self.ctx.ncalcs))
        print((self.ctx.ncalc))
        estimated_jobs = self.ctx.ncalc*wf_dict.get('points', 5)
        joblimit = wf_dict.get('joblimit', 90)
        self.ctx.eos_run_steps = 1
        self.ctx.eos_steps_done = 0
        self.ctx.minindex = 0
        self.ctx.maxindex = self.ctx.ncalc -1
        self.ctx.eos_max_perstep = 10000 # init

        if estimated_jobs >= joblimit:
            self.ctx.eos_run_steps = estimated_jobs/joblimit + 1
            self.ctx.eos_max_perstep = joblimit/wf_dict.get('points', 5)
            # TODO be carefull if is not a divisor... of joblimit
            self.ctx.maxindex = 0 # will be set later self.ctx.eos_max_perstep

        self.report('{} {}'.format(self.ctx.ncalc, self.ctx.eos_max_perstep))
        self.report('Estimated fleur scfs to run {}, running in {} steps.'
                    ''.format(estimated_jobs, self.ctx.eos_run_steps))

    def get_calcs_from_groups(self):
        """
        Extract the crystal structures and parameter data nodes from the given
        groups and create calculation 'pairs' (stru, para).
        """
        wf_dict = self.inputs.wf_parameters.get_dict()
        #get all delta structure
        str_gr = wf_dict.get('struc_group', 'delta')

        try:
            group_pk = int(str_gr)
        except ValueError:
            group_pk = None
            group_name = str_gr

        if group_pk is not None:
            try:
                str_group = Group(label=group_pk)
            except NotExistent:
                str_group = None
                message = ('You have to provide a valid pk for a Group of'
                          'structures or a Group name. Wf_para key: "struc_group".'
                          'given pk= {} is not a valid group'
                          '(or is your group name integer?)'.format(group_pk))
                #print(message)
                self.report(message)
                self.abort_nowait('I abort, because I have no structures to calculate ...')
        else:
            try:
                str_group = Group.get_from_string(group_name)
            except NotExistent:
                str_group = None
                message = ('You have to provide a valid pk for a Group of'
                          'structures or a Group name. Wf_para key: "struc_group".'
                          'given group name= {} is not a valid group'
                          '(or is your group name integer?)'.format(group_name))
                #print(message)
                self.report(message)
                self.abort_nowait('I abort, because I have no structures to calculate ...')


        #get all delta parameters
        para_gr = wf_dict.get('para_group', 'delta')

        if not para_gr:
            #waring use defauls
            message = 'COMMENT: I did recieve "para_group=None" as input. I will use inpgen defaults'
            self.report(message)

        try:
            group_pk = int(para_gr )
        except ValueError:
            group_pk = None
            group_name = para_gr

        if group_pk is not None:
            try:
                para_group = Group(label=group_pk)
            except NotExistent:
                para_group = None
                message = ('ERROR: You have to provide a valid pk for a Group of'
                          'parameters or a Group name (or use None for inpgen defaults). Wf_para key: "para_group".'
                          'given pk= {} is not a valid group'
                          '(or is your group name integer?)'.format(group_pk))
                #print(message)
                self.report(message)
                self.abort_nowait('ERROR: I abort, because I have no paremeters to calculate and '
                                  'I guess you did not want to use the inpgen default...')
        else:
            try:
                para_group = Group.get_from_string(group_name)
            except NotExistent:
                para_group = None
                message = ('ERROR: You have to provide a valid pk for a Group of'
                          'parameters or a Group name (or use None for inpgen defaults). Wf_para key: "struc_group".'
                          'given group name= {} is not a valid group'
                          '(or is your group name integer?)'.format(group_name))
                #print(message)
                self.report(message)
                self.abort_nowait('ERROR: I abort, because I have no paremeters to calculate and '
                                  'I guess you did not want to use the inpgen default...')

        # creating calculation pairs (structure, parameters)

        para_nodesi = para_group.nodes
        para_nodes = []

        for para in para_nodesi:
            para_nodes.append(para)
        #print para_nodes
        n_para = len(para_nodes)
        stru_nodes = str_group.nodes
        n_stru = len(stru_nodes)
        if n_para != n_stru:
            message = ('COMMENT: You did not provide the same number of parameter'
                       'nodes as structure nodes. Is this wanted? npara={} nstru={}'.format(n_para, n_stru))
            self.report(message)
        calcs = []
        for struc in stru_nodes:
            para = get_paranode(struc, para_nodes)
            #if para:
            calcs.append((struc, para))
            #else:
            #    calcs.append((struc))
        #pprint(calcs[:20])
        self.ctx.calcs_to_run = calcs
        self.ctx.ncalc = len(calcs)
        return

    #def calculations_left_torun(self):
    #    """
    #    Checks if there are still some equations of states to run
    #    """
    #    calculations_left = True
    #    self.ctx.last_step = False
    #
    #    if self.ctx.eos_steps_done == self.ctx.eos_run_steps:
    #        calculations_left = False
    #    if (self.ctx.eos_steps_done + 1) == self.ctx.eos_run_steps:
    #        self.ctx.last_step = True
    #
    #    return calculations_left



    def run_eos(self):
        """
        Run the equation of states for all delta structures with their parameters
        """
        #if self.ctx.last_step:
        #    self.ctx.maxindex = None
        #else:
        #    self.ctx.maxindex = self.ctx.maxindex + self.ctx.eos_max_perstep

        #self.report('Submitting eqaution of states part {} out of {}, from {} to {}'
        #            ''.format(self.ctx.eos_steps_done, self.ctx.eos_run_steps,
        #                      self.ctx.minindex, self.ctx.maxindex))

        eos_results = {}
        inputs = self.get_inputs_eos()


        #print(self.ctx.minindex)
        #print(self.ctx.maxindex)

        for struc, para in self.ctx.calcs_to_run:#[:4]self.ctx.minindex:self.ctx.maxindex]:#0:0]:#
            #print para
            formula = struc.get_formula()
            label = '|delta_wc|eos|{}'.format(formula)
            description = '|delta| fleur_eos_wc on {}'.format(formula)
            if para:
                eos_future = submit(FleurEosWorkChain,
                                wf_parameters=inputs['wc_eos_para'], structure=struc, options=inputs['options'],
                                calc_parameters=para, inpgen=inputs['inpgen'], fleur=inputs['fleur'],
                                label=label, description=description)
            else: # TODO: run eos_wc_simple
                eos_future = submit(FleurEosWorkChain,
                                wf_parameters=inputs['wc_eos_para'], structure=struc, options=inputs['options'],
                                inpgen=inputs['inpgen'], fleur=inputs['fleur'],
                                label=label, description=description)
            self.report('launching fleur_eos_wc<{}> on structure {} with parameter {}'
                        ''.format(eos_future.pid, struc.pk, para.pk))
            label = formula
            self.ctx.labels.append(label)
            eos_results[label] = eos_future

        #self.ctx.eos_steps_done = self.ctx.eos_steps_done + 1
        #self.ctx.minindex = self.ctx.maxindex


        return ToContext(**eos_results)

        '''
        # with run
        eos_results = {}
        inputs = self.get_inputs_eos()


        for struc, para in self.ctx.calcs_to_run[:]:
            print para
            formula = struc.get_formula()
            if para:
                #print('here')
                eos_future = fleur_eos_wc.run(
                                wf_parameters=inputs['wc_eos_para'], structure=struc,
                                calc_parameters=para, inpgen=inputs['inpgen'], fleur=inputs['fleur'])
                #fleur_eos_wc.run(#
            else:
                self.report('INFO: default parameters for structure {}'.format(formula))
                eos_future = fleur_eos_wc.run(
                                wf_parameters=inputs['wc_eos_para'], structure=struc,
                                inpgen=inputs['inpgen'], fleur=inputs['fleur'])
                #fleur_eos_wc.run(#a
            #self.report('launching fleur_eos_wc<{}> on structure {} with parameter {}'
            #            ''.format(eos_future.pid, struc.pk, para.pk))
            label = formula
            self.ctx.labels.append(label)
            eos_results[label] = eos_future

        return ToContext(**eos_results)
        '''

    # To limit the troughput of 100 jobs, we create several run eos steps
    def get_inputs_eos(self):
        """
        get the inputs for a scf-cycle
        """
        inputs = {}
        # produce the inputs for a eos worklfow (collect here...)

        inputs['wc_eos_para'] = self.ctx.wc_eos_para
        #inputs['calc_parameters'] = self.inputs.calc_parameters
        inputs['inpgen'] = self.ctx.inputs_eos.get('inpgen')
        inputs['fleur'] = self.ctx.inputs_eos.get('fleur')
        inputs['options'] = self.ctx.inputs_eos.get('options')
        return inputs

    def extract_results_eos(self):
        """
        extract information out of the result nodes of the the eos workchains
        ran in the step before
        """

        self.ctx.all_results = {}
        self.ctx.all_succ = {}
        self.ctx.eos_uuids ={}
        outstr = ('''\
             Delta calculation FLEUR {} (AiiDA wc).

             Crystal \t V0 \t \t  B0 \t \t  BP [A^3/at] \t [GPa] \t \t [--] \n
             '''.format(self.ctx.inputs_eos.get('fleur')))
        filename = 'delta_wc_{}.out'.format(self.ctx.own_pk)
        outfile = open(filename, 'w')
        outfile.write(outstr)
        outfile.close()
        outstr = ''
        for label in self.ctx.labels:
            eos_res = self.ctx[label]
            #print(calc)
            outpara1 = eos_res.get_outputs_dict()
            #print outpara1
            try:
                outpara= outpara1['output_eos_wc_para'].get_dict()
            except KeyError:
                self.report('ERROR: Eos wc for element: {} failed. I retrieved {} '
                            'I skip the results retrieval for that element.'.format(label, eos_res))
                continue
            eos_succ = outpara.get('successful', False)
            if not eos_succ:
                #maybe do something else here (exclude point and write a warning or so, or error treatment)
                self.ctx.successful = False

            natoms = outpara.get('natoms', None)
            gs_vol = outpara.get('volume_gs', None)
            bm = outpara.get('bulk_modulus', None)
            #bm_u = outpara.get('bulk_modulus_units', 'GPa')
            dbm = outpara.get('bulk_deriv', None)
            if natoms:
                gs_vol_pera = gs_vol/natoms
            else:
                gs_vol_pera = gs_vol

            element = label.translate(None, digits) # remove all numbers from string
            self.ctx.all_results[element] = [gs_vol_pera, bm, dbm]
            self.ctx.all_succ[element] = eos_succ
            self.ctx.eos_uuids[element] = eos_res.get_inputs()[0].uuid

            outstr = outstr + '{} \t {:.5f} \t {:.5f} \t {:.5f} \n'.format(element, gs_vol_pera, bm, dbm)
            #write inside the loop to have at least partially results...
            #outfile = open('delta_wc.out', 'a')
            #outstr = '{} \t {:.5f} \t {:.5f} \t {:.5f} \n'.format(element, gs_vol_pera, bm, dbm)
            #outfile.write(outstr)
            #outfile.close()
        # produce a single file
        # maybe put in try(or write in a certain place where is sure that you have the permissions)
        #outfile = open('delta_wc.out', 'w')
        outfile = open(filename, 'a') # for testing purposes
        outfile.write(outstr)

        outfile.close()

        self.ctx.outfilepath = os.path.abspath(outfile.name)


    def calculate_delta(self):
        """
        Execute here the script to calculate a delta factor
        """
        pass

    def return_results(self):
        """
        return the results of the calculations
        """

        # log some stuff in report

        # a text file should be written and stored as single file data and
        #parameter data node in the database

        #produce a single file data with all the numbers

        all_res = self.ctx.all_results
        self.report('all_res : {}'.format(all_res))
        #print all_res
        bm_dic = {}
        bmd_dic = {}
        vol_dic = {}

        for elem,val in six.iteritems(all_res):
            #print elem
            vol_dic[elem] = val[0]
            bm_dic[elem] = val[1]
            bmd_dic[elem] = val[2]

        outputnode_dict ={}


        outputnode_dict['workflow_name'] = self.__class__.__name__
        outputnode_dict['warnings'] = self.ctx.warnings
        outputnode_dict['successful'] = self.ctx.successful
        outputnode_dict['eos_uuids'] = self.ctx.eos_uuids
        outputnode_dict['eos_success'] = self.ctx.all_succ
        outputnode_dict['bulk_modulus'] = bm_dic
        outputnode_dict['bulk_modulus_units'] = 'GPa'
        outputnode_dict['bulk_modulus_dev'] = bmd_dic
        outputnode_dict['volumes'] = vol_dic
        outputnode_dict['volumes_units'] = 'A^3/per atom'
        outputnode_dict['delta_factor'] = {'Wien2K' : '', 'Fleur_026' : ''}

        #outputnode = Dict(dict=outputnode_dict)

        if self.ctx.successful:
            self.report('INFO: Done, delta worklfow complete')
            #print 'Done, delta worklfow complete'
        else:
            self.report('INFO: Done, but something went wrong.... Properly some '
                        'individual eos workchain failed. Check the log.')
            #print('Done, but something went wrong.... Properly some '
            #            'individual eos workchain failed. Check the log.')

        delta_file = SinglefileData.filename = self.ctx.outfilepath

        print(delta_file)

        # output must be aiida Data types.
        outnodedict = {}
        outnode = Dict(dict=outputnode_dict)
        outnodedict['results_node'] = outnode
        for label in self.ctx.labels:
            eos_res = self.ctx[label]
            #print(calc)
            outpara1 = eos_res.get_outputs_dict()
            #print outpara1
            try:
                outpara = outpara1['output_eos_wc_para']
            except KeyError:
                #self.report('ERROR: Eos wc for element: {} failed. I retrieved {} '
                #            'I skip the results retrieval for that element.'.format(label, eos_res))
                continue
            outnodedict[label] = outpara

        outputnode = create_delta_result_node(**outnodedict)

        outdict = {}
        outdict['output_delta_wc_para'] = outputnode.get('output_delta_wc_para')
        #outdict['delta_file'] = delta_file
        #print outdict
        for link_name, node in six.iteritems(outdict):
            self.out(link_name, node)
'''
if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description='SCF with FLEUR. workflow to'
                 ' converge the chargedensity and optional the total energy.')
    parser.add_argument('--wf_para', type=Dict, dest='wf_parameters',
                        help='The pseudopotential family', required=False)
    parser.add_argument('--structure', type=StructureData, dest='structure',
                        help='The crystal structure node', required=False)
    parser.add_argument('--calc_para', type=Dict, dest='calc_parameters',
                        help='Parameters for the FLEUR calculation', required=False)
    parser.add_argument('--fleurinp', type=FleurinpData, dest='fleurinp',
                        help='FleurinpData from which to run the FLEUR calculation', required=False)
    parser.add_argument('--remote', type=RemoteData, dest='remote_data',
                        help=('Remote Data of older FLEUR calculation, '
                        'from which files will be copied (mixing_history ...)'), required=False)
    parser.add_argument('--inpgen', type=Code, dest='inpgen',
                        help='The inpgen code node to use', required=False)
    parser.add_argument('--fleur', type=Code, dest='fleur',
                        help='The FLEUR code node to use', required=True)

    args = parser.parse_args()
    res = fleur_scf_wc.run(wf_parameters=args.wf_parameters,
                                structure=args.structure,
                                calc_parameters=args.calc_parameters,
                                fleurinp=args.fleurinp,
                                remote_data=args.remote_data,
                                inpgen = args.inpgen,
                                fleur=args.fleur)
'''
@cf
def create_delta_result_node(**kwargs):#*args):
    """
    This is a pseudo wf, to create the rigth graph structure of AiiDA.
    This wokfunction will create the output node in the database.
    It also connects the output_node to all nodes the information commes from.
    So far it is just also parsed in as argument, because so far we are to lazy
    to put most of the code overworked from return_results in here.

    """
    outdict = {}
    outpara =  kwargs.get('results_node', {})
    outdict['output_delta_wc_para'] = outpara.clone()
    # copy, because we rather produce the same node twice then have a circle in the database for now...
    #output_para = args[0]
    #return {'output_eos_wc_para'}
    return outdict


def get_paranode(struc, para_nodes):
    """
    find out if a parameter node for a structure is in para_nodes
    (currently very creedy, but lists are small (100x100) but maybe reduce database accesses)
    """

    suuid = struc.uuid
    formula = struc.get_formula()
    element = formula.translate(None, digits)
    #print para_nodes
    for para in para_nodes:
        struc_uuid = para.get_extra('struc_uuid', None)
        para_form = para.get_extra('formula', None)
        para_ele = para.get_extra('element', None)
        if suuid == struc_uuid:
            return para
        elif formula == para_form:
            return para
        elif element == para_ele:
            return para
        elif element == para_form:
            return para
        else:
            pass
            #Do something else (test if parameters for a certain element are there)
            #....
    # we found no parameter node for the given structure therefore return none
    return None

def write_delta_file(result_dict):
    pass




# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
"""
This module contains the parser for a FLEUR calculation and methods for parsing
different files produced by FLEUR.

Please implement file parsing routines that they can be executed from outside
the parser. Makes testing and portability easier.
"""
# TODO: move methods to utils, xml or other
# TODO: warnings
from __future__ import absolute_import
import os
import re
import json
from datetime import date
from lxml import etree

from aiida.parsers import Parser
from aiida.orm import Dict, BandsData
from aiida.common.exceptions import NotExistent


class FleurParser(Parser):
    """
    This class is the implementation of the Parser class for FLEUR.
    It parses the FLEUR output if the calculation was successful,
    i.e checks if all files are there that should be and their condition.
    Then it parses the out.xml file and returns a (simple) parameterData node
    with the results of the last iteration.
    Other files (DOS.x, bands.x, relax.xml, ...) are also parsed if they are retrieved.
    """

    _setting_key = 'parser_options'

    def get_linkname_outparams_complex(self):
        """
        Returns the name of the link to the output_complex
        Node contains the Fleur output in a rather complex dictionary.
        """
        return 'output_complex'

    def get_linkname_outparams(self):
        """
        Returns the name of the link to the output_complex
        Node contains the Fleur output in a rather complex dictionary.
        """
        return 'output_parameters'

    def parse(self, **kwargs):
        """
        Receives in input a dictionary of retrieved nodes.
        Does all the logic here. Checks presents of files.
        Calls routines to parse them and returns parameter nodes and success.

        :return successful: Bool, if overall parsing was successful or not
        :return new_nodes_list: list of tuples of two (linkname, Dataobject),
                                nodes to be stored by AiiDA

        """

        ####### init some variables ######

        # these files should be at least present after success of a Fleur run
        calc = self.node
        FleurCalculation = calc.process_class

        # this files should be retrieved
        should_retrieve = calc.get_attribute('retrieve_list')

        has_xml_outfile = False
        has_dos_file = False
        has_bands_file = False
        has_relax_file = False

        dos_file = None
        band_file = None

        ######### Check presence of files ######

        # select the folder object
        # Check that the retrieved folder is there
        try:
            output_folder = self.retrieved
        except NotExistent:
            self.logger.error("No retrieved folder found")
            return self.exit_codes.ERROR_NO_RETRIEVED_FOLDER

        # check what is inside the folder
        list_of_files = output_folder.list_object_names()
        self.logger.info("file list {}".format(list_of_files))

        # has output xml file, otherwise error
        if FleurCalculation._OUTXML_FILE_NAME not in list_of_files:
            self.logger.error(
                "XML out not found '{}'".format(FleurCalculation._OUTXML_FILE_NAME))
            return self.exit_codes.ERROR_NO_OUTXML
        else:
            has_xml_outfile = True

        # check if all files expected are there for the calculation
        for file in should_retrieve:
            if file not in list_of_files:
                self.logger.warning(
                    "'{}' file not found in retrived folder, it"
                    " was probably not created by fleur".format(file))

        # check if something was written to the error file
        if FleurCalculation._ERROR_FILE_NAME in list_of_files:
            errorfile = FleurCalculation._ERROR_FILE_NAME
            # read
            try:
                with output_folder.open(errorfile, 'r') as efile:
                    error_file_lines = efile.read()  # Note: read(), not readlines()
            except IOError:
                self.logger.error(
                    "Failed to open error file: {}.".format(errorfile))
                return self.exit_codes.ERROR_OPENING_OUTPUTS

            if error_file_lines:

                if 'Run finished successfully' not in error_file_lines:
                    self.logger.warning(
                        'The following was written into std error and piped to {}'
                        ' : \n {}'.format(errorfile, error_file_lines))
                    self.logger.error('FLEUR calculation did not finish'
                                      ' successfully.')

                    mpiprocs = self.node.get_attribute('resources').get(
                        'num_mpiprocs_per_machine', 1)

                    with output_folder.open('out.xml', 'r') as out_file: #lazy out.xml parsing
                        outlines = out_file.readline()
                        line_avail = re.findall(r'<mem memoryPerNode ="\d+', outlines)[0]
                        mem_kb_avail =int(re.findall(r'\d+', line_avail)[0])

                    if FleurCalculation._USAGE_DATA_FILE_NAME in list_of_files:
                        with output_folder.open(FleurCalculation._USAGE_FILE_NAME, 'r') as us_file:
                            usage = json.load(us_file)
                        kb_used = usage['VmPeak']
                    else:
                        try:
                            line_used = re.findall(r'used.+', error_file_lines)[0]
                            kb_used = int(re.findall(r'\d+', line_used)[2])
                        except IndexError:
                            kb_used = 0.0
                            self.logger.info('Did not manage to find memory usage info.')

                    if kb_used * mpiprocs / mem_kb_avail > 0.93:
                        return self.exit_codes.ERROR_NOT_ENOUGH_MEMORY
                    elif 'Atom spills out into vacuum during relaxation' in error_file_lines:
                        return self.exit_codes.ERROR_VACUUM_SPILL_RELAX
                    elif 'Error checking M.T. radii' in error_file_lines:
                        return self.exit_codes.ERROR_MT_RADII
                    elif 'Overlapping MT-spheres during relaxation: ' in error_file_lines:
                        over_indices = re.findall(
                            r'relaxation: +\S+ +\S+ +\S+', error_file_lines)[0].split()[1:]
                        error_params = {'error_name': 'MT_OVERLAP_RELAX',
                                        'description': ('This output node contains information'
                                                        'about FLEUR error'),
                                        'overlapped_indices': over_indices[:2],
                                        'overlaping_value': over_indices[2]}
                        link_name = self.get_linkname_outparams()
                        error_params = Dict(dict=error_params)
                        self.out('error_params', error_params)
                        return self.exit_codes.ERROR_MT_RADII_RELAX
                    else:
                        return self.exit_codes.ERROR_FLEUR_CALC_FAILED

        if FleurCalculation._DOS_FILE_NAME in list_of_files:
            has_dos = True
        if FleurCalculation._BAND_FILE_NAME in list_of_files:
            has_bands = True

        # if a relax.xml was retrieved
        if FleurCalculation._RELAX_FILE_NAME in list_of_files:
            self.logger.info("relax.xml file found in retrieved folder")
            has_relax_file = True

        ####### Parse the files ########

        if has_xml_outfile:
            # open output file
            outxmlfile_opened = output_folder.open(FleurCalculation._OUTXML_FILE_NAME, 'r')
            simpledata, complexdata, parser_info, success = parse_xmlout_file(outxmlfile_opened)
            outxmlfile_opened.close()

            # Call routines for output node creation
            if not success:
                self.logger.error(
                    'Parsing of XML output file was not successfull.')
                return self.exit_codes.ERROR_XMLOUT_PARSING_FAILED
            elif simpledata:
                outputdata = dict(list(simpledata.items()) +
                                  list(parser_info.items()))
                outputdata['CalcJob_uuid'] = self.node.uuid
                outxml_params = Dict(dict=outputdata)
                link_name = self.get_linkname_outparams()
                self.out(link_name, outxml_params)
            elif complexdata:
                parameter_data = dict(
                    list(complexdata.items()) + list(parser_info.items()))
                parameter_data['CalcJob_uuid'] = self.node.uuid
                outxml_params_complex = Dict(dict=parameter_data)
                link_name = self.get_linkname_outparams_complex()
                self.out(link_name, outxml_params_complex)
            else:
                self.logger.error(
                    "Something went wrong, neither simpledata nor complexdata found")
                parameter_data = dict(list(parser_info.items()))
                outxml_params = Dict(dict=parameter_data)
                link_name = self.get_linkname_outparams()
                self.out(link_name, outxml_params)

        # optional parse other files
        # DOS
        if has_dos_file:
            dos_file = FleurCalculation._DOS_FILE_NAME
            # if dos_file is not None:
            try:
                with output_folder.open(dos_file, 'r') as dosf:
                    dos_lines = dosf.read()  # Note: read() and not readlines()
            except IOError:
                self.logger.error(
                    'Failed to open DOS file: {}.'.format(dos_file))
                return self.exit_codes.ERROR_OPENING_OUTPUTS
            dos_data = parse_dos_file(dos_lines)  # , number_of_atom_types)

        # Bands
        if has_bands_file:
            # TODO: be carefull there might be two files.
            band_file = FleurCalculation._BAND_FILE_NAME

            # if band_file is not None:
            try:
                with output_folder.open(band_file, 'r') as bandf:
                    bands_lines = bandf.read()  # Note: read() and not readlines()
            except IOError:
                self.logger.error("Failed to open bandstructure file: {}."
                                  "".format(band_file))
                return self.exit_codes.ERROR_OPENING_OUTPUTS
            bands_data = parse_bands_file(bands_lines)

        if has_relax_file:
            relax_name = FleurCalculation._RELAX_FILE_NAME
            try:
                fleurinp = calc.inputs.fleurinpdata
            except NotExistent:
                old_relax_text = ''
            else:
                if relax_name in fleurinp.list_object_names():
                    with fleurinp.open(relax_name, 'r') as rlx:
                        old_relax_text = rlx.read()
                else:
                    old_relax_text = ''

            # dummy comparison between old and new relax
            with output_folder.open(relax_name, 'r') as rlx:
                new_relax_text = rlx.read()
                if new_relax_text != old_relax_text:
                    try:
                        relax_dict = parse_relax_file(rlx)
                    except etree.XMLSyntaxError:
                        return self.exit_codes.ERROR_RELAX_PARSING_FAILED
                    self.out('relax_parameters', relax_dict)


def parse_xmlout_file(outxmlfile):
    """
    Parses the out.xml file of a FLEUR calculation
    Receives as input the absolute path to the xml output file

    :param outxmlfile: path to out.xml file

    :returns xml_data_dict: a simple dictionary (QE output like)
                            with parsed data

    """
    from lxml import etree

    global parser_info_out

    parser_info_out = {'parser_warnings': [], 'unparsed': []}
    parser_version = '0.2beta'
    parser_info_out['parser_info'] = 'AiiDA Fleur Parser v{}'.format(
        parser_version)
    #parsed_data = {}

    successful = True
    outfile_broken = False
    parse_xml = True
    parser = etree.XMLParser(recover=False)

    try:
        tree = etree.parse(outxmlfile, parser)
    except etree.XMLSyntaxError:
        outfile_broken = True
        parser_info_out['parser_warnings'].append(
            'The out.xml file is broken I try to repair it.')

    if outfile_broken:
        # repair xmlfile and try to parse what is possible.
        parser = etree.XMLParser(recover=True)
        try:
            tree = etree.parse(outxmlfile, parser)
        except etree.XMLSyntaxError:
            parser_info_out['parser_warnings'].append(
                'Skipping the parsing of the xml file. '
                'Repairing was not possible.')
            parse_xml = False
            successful = False

    def parse_simplexmlout_file(root, outfile_broken):
        """
        Parses the xml.out file of a Fleur calculation
        Receives in input the root of an xmltree of the xml output file

        :param root: root node of etree of out.xml file
        :param outfile_broken: a boolen that indicates if an out.xml has a broken last iteration
                               output

        :returns xml_data_dict: a simple dictionary (QE output like)
                                with parsed data
        """

        ### all xpath used. (maintain this) ###
        iteration_xpath = '/fleurOutput/scfLoop/iteration'
        magnetism_xpath = '/fleurOutput/inputData/calculationSetup/magnetism'

        relPos_xpath = '/fleurOutput/inputData/atomGroups/atomGroup/relPos'
        absPos_xpath = '/fleurOutput/inputData/atomGroups/atomGroup/absPos'
        filmPos_xpath = '/fleurOutput/inputData/atomGroups/atomGroup/filmPos'

        atomstypes_xpath = '/fleurOutput/inputData/atomGroups/atomGroup'
        symmetries_xpath = '/fleurOutput/inputData/cell/symmetryOperations/symOp'
        kpoints_xpath = '/fleurOutput/inputData/calculationSetup/bzIntegration/kPointList/kPoint'
        species_xpath = '/fleurOutput/inputData/atomSpecies'

        # input parameters
        creator_name_xpath = 'programVersion/@version'
        output_version_xpath = '/fleurOutput/@fleurOutputVersion'
        creator_target_architecture_xpath = 'programVersion/targetComputerArchitectures/text()'
        creator_target_structure_xpath = 'programVersion/targetStructureClass/text()'
        precision_xpath = 'programVersion/precision/@type'

        title_xpath = '/fleurOutput/inputData/comment/text()'
        kmax_xpath = 'calculationSetup/cutoffs'
        gmax_xpath = 'calculationSetup/cutoffs'
        mixing_xpath = 'calculationSetup/scfLoop'
        number_of_bands_xpath = 'calculationSetup/cutoffs'
        spin_orbit_calculation = 'calculationSetup/soc'
        smearing_energy_xpath = 'calculationSetup/bzIntegration/@fermiSmearingEnergy'
        jspin_name = 'jspins'
        l_f_xpath = '/fleurOutput/inputData/calculationSetup/geometryOptimization/@l_f'

        # timing
        start_time_xpath = '/fleurOutput/startDateAndTime/@time'
        end_time_xpath = '/fleurOutput/endDateAndTime/@time'
        start_date_xpath = '/fleurOutput/startDateAndTime/@date'
        end_date_xpath = '/fleurOutput/endDateAndTime/@date'

        ###########

        # get all iterations in out.xml file
        iteration_nodes = eval_xpath2(root, iteration_xpath)
        n_iters = len(iteration_nodes)
        data_exists = True

        # parse only last stable interation
        # (if modes (dos and co) maybe parse anyway if broken?)
        if outfile_broken and (n_iters >= 2):
            iteration_to_parse = iteration_nodes[-2]
            parser_info_out['last_iteration_parsed'] = n_iters-2
        elif outfile_broken and (n_iters == 1):
            iteration_to_parse = iteration_nodes[0]
            parser_info_out['last_iteration_parsed'] = n_iters
        elif not outfile_broken and (n_iters >= 1):
            iteration_to_parse = iteration_nodes[-1]
        else:  # there was no iteration found.
            # only the starting charge density could be generated
            parser_info_out['parser_warnings'].append(
                'There was no iteration found in the outfile, either just a '
                'starting density was generated or something went wrong.')
            data_exists = False
            iteration_to_parse = None

        # for getting the fleur modes use fleurinp methods
        spin = get_xml_attribute(eval_xpath(root, magnetism_xpath), jspin_name)

        if spin:
            fleurmode = {'jspin': int(spin)}
        else:
            fleurmode = {'jspin': 1}

        relax = eval_xpath(root, l_f_xpath)
        fleurmode['relax'] = relax == 'T'

        if data_exists:
            simple_data = parse_simple_outnode(iteration_to_parse, fleurmode)
        else:
            simple_data = {}

        # TODO: in the future add here the warnings returned from parse_simple_outnode
        # Currently Fleur warnings an errors are not written to the out.xml
        # should they be lists or dicts?
        warnings = {'info': {}, 'debug': {}, 'warning': {}, 'error': {}}

        simple_data['number_of_atoms'] = (len(eval_xpath2(root, relPos_xpath))
                                          + len(eval_xpath2(root, absPos_xpath))
                                          + len(eval_xpath2(root, filmPos_xpath)))
        simple_data['number_of_atom_types'] = len(eval_xpath2(root, atomstypes_xpath))
        simple_data['number_of_iterations'] = n_iters
        simple_data['number_of_symmetries'] = len(eval_xpath2(root, symmetries_xpath))
        simple_data['number_of_species'] = len(eval_xpath2(root, species_xpath))
        simple_data['number_of_kpoints'] = len(eval_xpath2(root, kpoints_xpath))
        simple_data['number_of_spin_components'] = fleurmode['jspin']

        title = eval_xpath(root, title_xpath)
        if title:
            title = str(title).strip()
        simple_data['title'] = title
        simple_data['creator_name'] = eval_xpath(root, creator_name_xpath)
        simple_data['creator_target_architecture'] = eval_xpath(
            root, creator_target_architecture_xpath)
        simple_data['creator_target_structure'] = eval_xpath(
            root, creator_target_structure_xpath)
        simple_data['output_file_version'] = eval_xpath(
            root, output_version_xpath)

        # time
        # Maybe change the behavior if things could not be parsed...
        # Especially if file was broken, ie endtime it not there.
        starttime = eval_xpath(root, start_time_xpath)
        if starttime:
            starttimes = starttime.split(':')
        else:
            starttimes = [0, 0, 0]
            msg = 'Startime was unparsed, inp.xml prob not complete, do not believe the walltime!'
            if data_exists:
                parser_info_out['parser_warnings'].append(msg)

        endtime = eval_xpath(root, end_time_xpath)
        if endtime:
            endtimes = endtime.split(':')
        else:
            endtimes = [0, 0, 0]
            msg = 'Endtime was unparsed, inp.xml prob not complete, do not believe the walltime!'
            if data_exists:
                parser_info_out['parser_warnings'].append(msg)
        start_date = eval_xpath(root, start_date_xpath)
        end_date = eval_xpath(root, end_date_xpath)

        offset = 0
        if start_date != end_date:
            # date="2018/01/15", Can this fail? what happens if not there
            if start_date and end_date:
                date_sl = [int(ent) for ent in start_date.split('/')]
                date_el = [int(ent) for ent in end_date.split('/')]
                date_s = date(*date_sl)
                date_e = date(*date_el)
                diff = date_e - date_s
                offset = diff.days * 86400
        # ncores = 12 #TODO parse parallelization_Parameters
        time = offset + (int(endtimes[0])-int(starttimes[0]))*60*60 + (
            int(endtimes[1])-int(starttimes[1]))*60 + int(endtimes[2]) - int(starttimes[2])
        simple_data['walltime'] = time
        simple_data['walltime_units'] = 'seconds'
        #simple_data['core_hours'] = time*ncores*1.0/3600
        #simple_data['parallelization_Parameters'] = {'mpiPEs' : ncores}
        simple_data['start_date'] = {'date': start_date, 'time': starttime}
        simple_data['end_date'] = {'date': end_date, 'time': endtime}

        warnings['info'] = {}  # TODO
        warnings['debug'] = {}  # TODO
        warnings['warning'] = {}  # TODO
        warnings['error'] = {}  # TODO
        simple_data['warnings'] = warnings

        return simple_data

    # TODO find a way to import these from xml_util, but make the parser logger work...
    def eval_xpath(node, xpath):
        """
        Tries to evaluate an xpath expression. If it fails it logs it.

        :param node: root node of an etree
        :param xpath: xpath expression (relative, or absolute)
        :returns: text, attribute or a node list
        """
        try:
            return_value = node.xpath(xpath)
        except etree.XPathEvalError:
            parser_info_out['parser_warnings'].append(
                'There was a XpathEvalError on the xpath: {} \n Either it does '
                'not exist, or something is wrong with the expression.'
                ''.format(xpath))
            return []  # or rather None?
        if len(return_value) == 1:
            return return_value[0]
        else:
            return return_value

    def eval_xpath2(node, xpath):
        """
        Tries to evaluate an xpath expression. If it fails it logs it.
        It always return a list even if a single element was evaluated.

        :param node: root node of an etree
        :param xpath: xpath expression (relative, or absolute)
        :returns: a node list
        """
        try:
            return_value = node.xpath(xpath)
        except etree.XPathEvalError:
            parser_info_out['parser_warnings'].append(
                'There was a XpathEvalError on the xpath: {} \n Either it does '
                'not exist, or something is wrong with the expression.'
                ''.format(xpath))
            return []
        return return_value

    def get_xml_attribute(node, attributename):
        """
        Get an attribute value from a node.

        :param node: a node from etree
        :param attributename: a string with the attribute name.
        :returns: attributevalue or None
        """
        if etree.iselement(node):
            attrib_value = node.get(attributename)
            if attrib_value:
                return attrib_value
            else:
                parser_info_out['parser_warnings'].append(
                    'Tried to get attribute: "{}" from element {}.\n '
                    'I received "{}", maybe the attribute does not exist'
                    ''.format(attributename, node, attrib_value))
                return None
        else:  # something doesn't work here, some nodes get through here
            parser_info_out['parser_warnings'].append(
                'Can not get attributename: "{}" from node "{}", '
                'because node is not an element of etree.'
                ''.format(attributename, node))

            return None

    def convert_to_float(value_string):
        """
        Tries to make a float out of a string. If it can't it logs a warning
        and returns True or False if convertion worked or not.

        :param value_string: a string
        :returns value: the new float or value_string: the string given
        :returns: True if convertation was successfull, False otherwise
        """
        try:
            value = float(value_string)
        except TypeError:
            parser_info_out['parser_warnings'].append(
                'Could not convert: "{}" to float, TypeError'
                ''.format(value_string))
            return value_string, False
        except ValueError:
            parser_info_out['parser_warnings'].append(
                'Could not convert: "{}" to float, ValueError'
                ''.format(value_string))
            return value_string, False
        return value, True

    def convert_to_int(value_string):
        """
        Tries to make a int out of a string. If it can't it logs a warning
        and returns True or False if convertion worked or not.

        :param value_string: a string
        :returns value: the new int or value_string: the string given
        :returns: True or False
        """
        try:
            value = int(value_string)
        except TypeError:
            parser_info_out['parser_warnings'].append(
                'Could not convert: "{}" to int, TypeError'
                ''.format(value_string))
            return value_string, False
        except ValueError:
            parser_info_out['parser_warnings'].append(
                'Could not convert: "{}" to int, ValueError'
                ''.format(value_string))
            return value_string, False
        return value, True

    def convert_htr_to_ev(value):
        """
        Multiplies the value given with the Hartree factor (converts htr to eV)
        """
        htr = 27.21138602
        suc = False
        value_to_save, suc = convert_to_float(value)
        if suc:
            return value_to_save*htr
        else:
            return value

    def parse_simple_outnode(iteration_node, fleurmode):
        """
        Parses the data from the iteration given (usually last iteration)
        and some other data for the 'simple' output node.

        :param iteration_node: etree node of an interation
        :param Fleurmode: python dictionary, with all the modes,
                          which influence the parsing

        :returns simple_data: a python dictionary with all results
        """
        ###################################################
        ##########  all xpaths (maintain this) ############
        # (specifies where to find things in the out.xml) #


        # density
        densityconvergence_xpath = 'densityConvergence'
        chargedensity_xpath = 'densityConvergence/chargeDensity'
        overallchargedensity_xpath = 'densityConvergence/overallChargeDensity'
        spindensity_xpath = 'densityConvergence/spinDensity'

        bandgap_xpath = 'bandgap'
        fermi_energy_xpath = 'FermiEnergy'

        # magnetic moments
        magnetic_moments_in_mtpheres_xpath = 'magneticMomentsInMTSpheres'
        magneticmoment_xpath = 'magneticMomentsInMTSpheres/magneticMoment'

        magneticmoments_xpath = 'magneticMomentsInMTSpheres/magneticMoment/@moment'
        magneticmoments_spinupcharge_xpath = 'magneticMomentsInMTSpheres/magneticMoment/@spinUpCharge'
        magneticmoments_spindowncharge_xpath = 'magneticMomentsInMTSpheres/magneticMoment/@spinDownCharge'

        orbmagnetic_moments_in_mtpheres_xpath = 'orbitalMagneticMomentsInMTSpheres'
        orbmagneticmoment_xpath = 'orbitalMagneticMomentsInMTSpheres/orbMagMoment'

        orbmagneticmoments_xpath = 'orbitalMagneticMomentsInMTSpheres/orbMagMoment/@moment'
        orbmagneticmoments_spinupcharge_xpath = 'orbitalMagneticMomentsInMTSpheres/orbMagMoment/@spinUpCharge'
        orbmagneticmoments_spindowncharge_xpath = 'orbitalMagneticMomentsInMTSpheres/orbMagMoment/@spinDownCharge'

        mae_force_theta_xpath = 'Forcetheorem_MAE/Angle/@theta'
        mae_force_phi_xpath = 'Forcetheorem_MAE/Angle/@phi'
        mae_force_evsum_xpath = 'Forcetheorem_MAE/Angle/@ev-sum'
        mae_force_energ_units_xpath = 'Forcetheorem_Loop_MAE/sumValenceSingleParticleEnergies/@units'

        spst_force_xpath = 'Forcetheorem_SSDISP/@qvectors'
        spst_force_q_xpath = 'Forcetheorem_SSDISP/Entry/@q'
        spst_force_evsum_xpath = 'Forcetheorem_SSDISP/Entry/@ev-sum'
        spst_force_energ_units_xpath = 'Forcetheorem_Loop_SSDISP/sumValenceSingleParticleEnergies/@units'

        dmi_force_xpath = 'Forcetheorem_DMI'
        dmi_force_q_xpath = 'Forcetheorem_DMI/Entry/@q'
        dmi_force_theta_xpath = 'Forcetheorem_DMI/Entry/@theta'
        dmi_force_phi_xpath = 'Forcetheorem_DMI/Entry/@phi'
        dmi_force_evsum_xpath = 'Forcetheorem_DMI/Entry/@ev-sum'
        dmi_force_angles_xpath = 'Forcetheorem_DMI/@Angles'
        dmi_force_qs_xpath = 'Forcetheorem_DMI/@qPoints'
        dmi_force_energ_units_xpath = 'Forcetheorem_Loop_DMI/sumValenceSingleParticleEnergies/@units'

        spinupcharge_name = 'spinUpCharge'
        spindowncharge_name = 'spinDownCharge'
        moment_name = 'moment'

        # all electron charges
        allelectronchages_xpath = ''

        a = 'total'
        b = 'interstitial'
        c = 'value'
        # energy
        totalenergy_xpath = 'totalEnergy'
        sumofeigenvalues_xpath = 'totalEnergy/sumOfEigenvalues'
        core_electrons_xpath = 'totalEnergy/sumOfEigenvalues/coreElectrons'
        valence_electrons_xpath = 'totalEnergy/sumOfEigenvalues/valenceElectrons'
        chargeden_xc_den_integral_xpath = 'totalEnergy/chargeDenXCDenIntegral'

        #free_energy_xpath = 'totalEnergy/freeEnergy'

        # forces
        forces_units_xpath = 'totalForcesOnRepresentativeAtoms'
        forces_total_xpath = 'totalForcesOnRepresentativeAtoms/forceTotal'

        #
        iteration_xpath = '.'

        units_name = 'units'
        value_name = 'value'
        distance_name = 'distance'

        overall_number_name = 'overallNumber'
        atomtype_name = 'atomType'

        # forces
        f_x_name = 'F_x'
        f_y_name = 'F_y'
        f_z_name = 'F_z'
        new_x_name = 'x'
        new_y_name = 'y'
        new_z_name = 'z'

        relPos_xpath = '/fleurOutput/inputData/atomGroups/atomGroup/relPos'
        absPos_xpath = '/fleurOutput/inputData/atomGroups/atomGroup/absPos'
        filmPos_xpath = '/fleurOutput/inputData/atomGroups/atomGroup/filmPos'
        atomstypes_xpath = '/fleurOutput/inputData/atomGroups/atomGroup'

        film_lat_xpath = '/fleurOutput/inputData/cell/filmLattice/bravaisMatrix/'
        bulk_lat_xpath = '/fleurOutput/inputData/cell/bulkLattice/bravaisMatrix/'

        kmax_xpath = '/fleurOutput/inputData/calculationSetup/cutoffs/@Kmax'

        ###################################################

        jspin = fleurmode['jspin']
        relax = fleurmode['relax']
        simple_data = {}

        def write_simple_outnode(value, value_type, value_name, dict_out):
            """
            Writes a value (int or float) in the simple data dict.
            if path does not exit it initializes it!

            If the value has not the type given,
            it will be logged in parser info instead of writing.

            :param value: value ti write
            :param value_type: a type of a value: 'float', 'int', 'str', 'list', 'list_floats',
                               'list_ints', list_list_floats
            :param value_name: a name of a value to be used in the dictionary
            """

            iteration_current_number_name = 'numberForCurrentRun'
            suc = False

            if value_type == 'float':
                value_to_save, suc = convert_to_float(value)
            elif value_type == 'int':
                value_to_save, suc = convert_to_int(value)
            elif value_type == 'str':
                suc = True
                value_to_save = value
                #value_to_save, suc = convert_to_str(value)
            elif value_type == 'list':
                suc = True
                value_to_save = value
            elif value_type == 'list_floats':
                value_to_save = []
                for val in value:
                    value_to_savet, suct = convert_to_float(val)
                    value_to_save.append(value_to_savet)
                suc = True  # TODO individual or common error message?
            elif value_type == 'list_ints':
                value_to_save = []
                for val in value:
                    value_to_savet, suct = convert_to_int(val)
                    value_to_save.append(value_to_savet)
                suc = True
            elif value_type == 'list_list_floats':
                value_to_save = []
                for val in value:
                    value_to_savet = []
                    for val1 in val:
                        value_to_savet1, suct = convert_to_float(val1)
                        value_to_savet.append(value_to_savet1)
                    value_to_save.append(value_to_savet)
                suc = True  # TODO individual or common error message?
            else:
                #self.logger.error('I dont know the type you gave me {}'.format(type))
                pass
                # TODO log error, self is not known here...
            if suc:
                dict_out[value_name] = value_to_save
            else:
                parser_info_out['unparsed'].append(
                    {value_name: value,
                     'iteration': get_xml_attribute(iteration_node, iteration_current_number_name)})

        if eval_xpath(iteration_node, mae_force_theta_xpath) != []:
            # extract MAE force theorem parameters
            mae_force_theta = eval_xpath2(
                iteration_node, mae_force_theta_xpath)
            write_simple_outnode(
                mae_force_theta, 'list_floats', 'mae_force_theta', simple_data)

            mae_force_evsum = eval_xpath2(
                iteration_node, mae_force_evsum_xpath)
            write_simple_outnode(
                mae_force_evsum, 'list_floats', 'mae_force_evSum', simple_data)

            mae_force_phi = eval_xpath2(iteration_node, mae_force_phi_xpath)
            write_simple_outnode(
                mae_force_phi, 'list_floats', 'mae_force_phi', simple_data)

            units_e = eval_xpath2(iteration_node, mae_force_energ_units_xpath)
            write_simple_outnode(
                units_e[0], 'str', 'energy_units', simple_data)
        elif eval_xpath(iteration_node, spst_force_xpath) != []:
            # extract Spin spiral dispersion force theorem parameters
            spst_force_q = eval_xpath2(iteration_node, spst_force_q_xpath)
            write_simple_outnode(
                spst_force_q, 'list_floats', 'spst_force_q', simple_data)

            spst_force_evsum = eval_xpath2(
                iteration_node, spst_force_evsum_xpath)
            write_simple_outnode(
                spst_force_evsum, 'list_floats', 'spst_force_evSum', simple_data)

            units_e = eval_xpath2(iteration_node, spst_force_energ_units_xpath)
            write_simple_outnode(
                units_e[0], 'str', 'energy_units', simple_data)
        elif eval_xpath(iteration_node, dmi_force_xpath) != []:
            # extract DMI force theorem parameters
            dmi_force_q = eval_xpath2(iteration_node, dmi_force_q_xpath)
            write_simple_outnode(
                dmi_force_q, 'list_ints', 'dmi_force_q', simple_data)

            dmi_force_evsum = eval_xpath2(
                iteration_node, dmi_force_evsum_xpath)
            write_simple_outnode(
                dmi_force_evsum, 'list_floats', 'dmi_force_evSum', simple_data)

            dmi_force_theta = eval_xpath2(
                iteration_node, dmi_force_theta_xpath)
            write_simple_outnode(
                dmi_force_theta, 'list_floats', 'dmi_force_theta', simple_data)

            dmi_force_phi = eval_xpath2(iteration_node, dmi_force_phi_xpath)
            write_simple_outnode(
                dmi_force_phi, 'list_floats', 'dmi_force_phi', simple_data)

            dmi_force_angles = eval_xpath(
                iteration_node, dmi_force_angles_xpath)
            write_simple_outnode(
                dmi_force_angles, 'int', 'dmi_force_angles', simple_data)

            dmi_force_qs = eval_xpath(iteration_node, dmi_force_qs_xpath)
            write_simple_outnode(
                dmi_force_qs, 'int', 'dmi_force_qs', simple_data)

            units_e = eval_xpath2(iteration_node, dmi_force_energ_units_xpath)
            write_simple_outnode(
                units_e[0], 'str', 'energy_units', simple_data)
        else:
            # total energy

            kmax_used = eval_xpath2(root, kmax_xpath)[0]
            write_simple_outnode(kmax_used, 'float', 'kmax', simple_data)

            units_e = get_xml_attribute(
                eval_xpath(iteration_node, totalenergy_xpath), units_name)
            write_simple_outnode(
                units_e, 'str', 'energy_hartree_units', simple_data)

            tE_htr = get_xml_attribute(
                eval_xpath(iteration_node, totalenergy_xpath), value_name)
            write_simple_outnode(
                tE_htr, 'float', 'energy_hartree', simple_data)

            write_simple_outnode(
                convert_htr_to_ev(tE_htr), 'float', 'energy', simple_data)
            write_simple_outnode('eV', 'str', 'energy_units', simple_data)

            sumofeigenvalues = get_xml_attribute(
                eval_xpath(iteration_node, sumofeigenvalues_xpath), value_name)
            write_simple_outnode(
                sumofeigenvalues, 'float', 'sum_of_eigenvalues', simple_data)

            coreElectrons = get_xml_attribute(
                eval_xpath(iteration_node, core_electrons_xpath), value_name)
            write_simple_outnode(
                coreElectrons, 'float', 'energy_core_electrons', simple_data)

            valenceElectrons = get_xml_attribute(
                eval_xpath(iteration_node, valence_electrons_xpath), value_name)
            write_simple_outnode(
                valenceElectrons, 'float', 'energy_valence_electrons', simple_data)

            ch_d_xc_d_inte = get_xml_attribute(
                eval_xpath(iteration_node, chargeden_xc_den_integral_xpath), value_name)
            write_simple_outnode(
                ch_d_xc_d_inte, 'float', 'charge_den_xc_den_integral', simple_data)

            # bandgap
            units_bandgap = get_xml_attribute(
                eval_xpath(iteration_node, bandgap_xpath), units_name)
            write_simple_outnode(units_bandgap, 'str',
                                 'bandgap_units', simple_data)

            bandgap = get_xml_attribute(
                eval_xpath(iteration_node, bandgap_xpath), value_name)
            write_simple_outnode(bandgap, 'float', 'bandgap', simple_data)

            # fermi
            fermi_energy = get_xml_attribute(
                eval_xpath(iteration_node, fermi_energy_xpath), value_name)
            write_simple_outnode(fermi_energy, 'float',
                                 'fermi_energy', simple_data)
            units_fermi_energy = get_xml_attribute(
                eval_xpath(iteration_node, fermi_energy_xpath), units_name)
            write_simple_outnode(
                units_fermi_energy, 'str', 'fermi_energy_units', simple_data)

            # density convergence
            units = get_xml_attribute(
                eval_xpath(iteration_node, densityconvergence_xpath), units_name)
            write_simple_outnode(
                units, 'str', 'density_convergence_units', simple_data)

            if jspin == 1:
                if not relax: # there are no charge densities written if relax
                    charge_density = get_xml_attribute(
                        eval_xpath(iteration_node, chargedensity_xpath), distance_name)
                    write_simple_outnode(
                        charge_density, 'float', 'charge_density', simple_data)

            elif jspin == 2:
                charge_densitys = eval_xpath2(
                    iteration_node, chargedensity_xpath)

                if not relax: # there are no charge densities written if relax
                    if charge_densitys:  # otherwise we get a keyerror if calculation failed.
                        charge_density1 = get_xml_attribute(
                            charge_densitys[0], distance_name)
                        charge_density2 = get_xml_attribute(
                            charge_densitys[1], distance_name)
                    else:  # Is non a problem?
                        charge_density1 = None
                        charge_density2 = None
                    write_simple_outnode(
                        charge_density1, 'float', 'charge_density1', simple_data)
                    write_simple_outnode(
                        charge_density2, 'float', 'charge_density2', simple_data)

                    spin_density = get_xml_attribute(
                        eval_xpath(iteration_node, spindensity_xpath), distance_name)
                    write_simple_outnode(
                        spin_density, 'float', 'spin_density', simple_data)

                    overall_charge_density = get_xml_attribute(
                        eval_xpath(iteration_node, overallchargedensity_xpath), distance_name)
                    write_simple_outnode(
                        overall_charge_density, 'float', 'overall_charge_density', simple_data)

                # magnetic moments            #TODO orbMag Moment
                m_units = get_xml_attribute(
                    eval_xpath(iteration_node, magnetic_moments_in_mtpheres_xpath), units_name)
                write_simple_outnode(
                    m_units, 'str', 'magnetic_moment_units', simple_data)
                write_simple_outnode(
                    m_units, 'str', 'orbital_magnetic_moment_units', simple_data)

                moments = eval_xpath(iteration_node, magneticmoments_xpath)
                write_simple_outnode(
                    moments, 'list_floats', 'magnetic_moments', simple_data)

                spinup = eval_xpath(
                    iteration_node, magneticmoments_spinupcharge_xpath)
                write_simple_outnode(
                    spinup, 'list_floats', 'magnetic_spin_up_charges', simple_data)

                spindown = eval_xpath(
                    iteration_node, magneticmoments_spindowncharge_xpath)
                write_simple_outnode(
                    spindown, 'list_floats', 'magnetic_spin_down_charges', simple_data)

                # orbital magnetic moments
                orbmoments = eval_xpath(
                    iteration_node, orbmagneticmoments_xpath)
                write_simple_outnode(
                    orbmoments, 'list_floats', 'orbital_magnetic_moments', simple_data)

                orbspinup = eval_xpath(
                    iteration_node, orbmagneticmoments_spinupcharge_xpath)
                write_simple_outnode(
                    orbspinup, 'list_floats', 'orbital_magnetic_spin_up_charges', simple_data)

                orbspindown = eval_xpath(
                    iteration_node, orbmagneticmoments_spindowncharge_xpath)
                write_simple_outnode(
                    orbspindown, 'list_floats', 'orbital_magnetic_spin_down_charges', simple_data)

                # TODO: atomtype dependence
                # moment = get_xml_attribute(
                #    eval_xpath(iteration_node, magneticmoment_xpath), moment_name)
                #write_simple_outnode(moment, 'float', 'magnetic_moment', simple_data)

                # spinup = get_xml_attribute(
                #    eval_xpath(iteration_node, magneticmoment_xpath), spinupcharge_name)
                #write_simple_outnode(spinup, 'float', 'spin_up_charge', simple_data)

                # spindown = get_xml_attribute(
                #    eval_xpath(iteration_node, magneticmoment_xpath), spindowncharge_name)
                #write_simple_outnode(spindown, 'float', 'spin_down_charge', simple_data)

                # Total charges, total magentic moment

            if relax:
                # check if it is a film or a bulk structure
                film = eval_xpath2(root, os.path.join(film_lat_xpath, 'row-1'))
                if film:
                    lat_path = film_lat_xpath
                    pos_attr = 'filmPos'
                else:
                    lat_path = bulk_lat_xpath
                    pos_attr = 'relPos'

                v_1 = eval_xpath(root, os.path.join(lat_path, 'row-1'))
                v_1 = [float(x) for x in v_1.text.split()]

                v_2 = eval_xpath(root, os.path.join(lat_path, 'row-2'))
                v_2 = [float(x) for x in v_2.text.split()]

                v_3 = eval_xpath(root, os.path.join(lat_path, 'row-3'))
                v_3 = [float(x) for x in v_3.text.split()]

                relax_brav_vectors = [v_1, v_2, v_3]

                atom_positions = []

                all_atoms = eval_xpath2(root, atomstypes_xpath)
                for a_type in all_atoms:
                    element = get_xml_attribute(a_type, 'species')[:2]
                    type_positions = eval_xpath2(a_type, pos_attr)
                    for pos in type_positions:
                        pos = [convert_frac(x) for x in pos.text.split()]
                        atom_positions.append([element]+pos)

                write_simple_outnode(
                    relax_brav_vectors, 'list', 'relax_brav_vectors', simple_data)
                write_simple_outnode(
                    atom_positions, 'list', 'relax_atom_positions', simple_data)
                write_simple_outnode(
                    str(bool(film)), 'str', 'film', simple_data)

            # total iterations
            number_of_iterations_total = get_xml_attribute(
                eval_xpath(iteration_node, iteration_xpath), overall_number_name)
            write_simple_outnode(
                number_of_iterations_total, 'int', 'number_of_iterations_total', simple_data)

            # forces atomtype dependend
            forces = eval_xpath2(iteration_node, forces_total_xpath)
            # length should be ntypes
            largest_force = -0.0
            for force in forces:
                atomtype, _ = convert_to_int(get_xml_attribute(force, atomtype_name))

                forces_unit = get_xml_attribute(
                    eval_xpath(iteration_node, forces_units_xpath), units_name)
                write_simple_outnode(forces_unit, 'str', 'force_units', simple_data)

                force_x = get_xml_attribute(force, f_x_name)
                write_simple_outnode(
                    force_x, 'float', 'force_x_type{}'.format(atomtype), simple_data)

                force_y = get_xml_attribute(force, f_y_name)
                write_simple_outnode(
                    force_y, 'float', 'force_y_type{}'.format(atomtype), simple_data)

                force_z = get_xml_attribute(force, f_z_name)
                write_simple_outnode(
                    force_z, 'float', 'force_z_type{}'.format(atomtype), simple_data)

                force_xf, suc1 = convert_to_float(force_x)
                force_yf, suc2 = convert_to_float(force_y)
                force_zf, suc3 = convert_to_float(force_z)

                if suc1 and suc2 and suc3:
                    if abs(force_xf) > largest_force:
                        largest_force = abs(force_xf)
                    if abs(force_yf) > largest_force:
                        largest_force = abs(force_yf)
                    if abs(force_zf) > largest_force:
                        largest_force = abs(force_zf)

                pos_x = get_xml_attribute(force, new_x_name)
                write_simple_outnode(
                    pos_x, 'float', 'abspos_x_type{}'.format(atomtype), simple_data)
                pos_y = get_xml_attribute(force, new_y_name)
                write_simple_outnode(
                    pos_y, 'float', 'abspos_y_type{}'.format(atomtype), simple_data)
                pos_z = get_xml_attribute(force, new_z_name)
                write_simple_outnode(
                    pos_z, 'float', 'abspos_z_type{}'.format(atomtype), simple_data)

            write_simple_outnode(
                largest_force, 'float', 'force_largest', simple_data)

        return simple_data

    if parse_xml:
        root = tree.getroot()
        simple_out = parse_simplexmlout_file(root, outfile_broken)
        #simple_out['outputfile_path'] = outxmlfile
        # TODO: parse complex out
        complex_out = {}  # parse_xmlout_file(root)
        return simple_out, complex_out, parser_info_out, successful
    else:
        return {}, {}, parser_info_out, successful


def parse_dos_file(dos_lines):  # , number_of_atom_types):
    """
    Parses the returned DOS.X files.
    Structure:
    (100(1x,e10.3))   e,totdos,interstitial,vac1,vac2,
    (at(i),i=1,ntype),((q(l,i),l=1,LMAX),i=1,ntype)
    where e is the energy in eV (= 1/27.2 htr) at(i) is the local DOS of a
    single atom of the i'th atom-type and q(l,i) is the l-resolved DOS at
    the i'th atom but has to be multiplied by the number of atoms of this type.

    :param dos_lines: string of the read in dos file
    :param number_of_atom_types: integer, number of atom types
    """
    # pass
    return 0


def parse_bands_file(bands_lines):
    '''
    Parses the returned bands.1 and bands.2 file and returns a complete
    bandsData object. bands.1 has the form: k value, energy

    :param bands_lines: string of the read in bands file
    '''
    # TODO: not finished
    # read bands out of file:
    nrows = 0  # get number of rows (known form number of atom types
    bands_values = []  # init an array of arrays nkpoint * ...
    bands_labels = []  # label for each row.

    # fill and correct fermi energy.
    bands_values = []

    # TODO: we need to get the cell from StructureData node
    # and KpointsData node from inpxml
    fleur_bands = BandsData()
    # fleur_bands.set_cell(cell)
    #fleur_bands.set_kpoints(kpoints, cartesian=True)
    fleur_bands.set_bands(bands=bands_values,
                          units='eV',
                          labels=bands_labels)

    for line in bands_lines:
        pass
    return fleur_bands


def parse_relax_file(rlx):
    """
    This function parsers relax.xml output file and
    returns a Dict containing all the data given there.
    """
    from aiida_fleur.tools.xml_util import eval_xpath2

    rlx.seek(0)
    tree = etree.parse(rlx)


    xpath_disp = '/relaxation/displacements/displace'
    xpath_energy = '/relaxation/relaxation-history/step/@energy'
    xpath_steps = '/relaxation/relaxation-history/step'

    root = tree.getroot()

    displacements = eval_xpath2(root, xpath_disp)
    float_displ = []
    for i in displacements:
        temp = [convert_frac(x) for x in i.text.split()]
        float_displ.append(temp)

    energies = eval_xpath2(root, xpath_energy)
    energies = [float(x) for x in energies]

    float_posforces = []
    iter_all = eval_xpath2(root, xpath_steps)
    for posf in iter_all:
        posforces = eval_xpath2(posf, 'posforce')
        temp2 = []
        for i in posforces:
            temp = [convert_frac(x) for x in i.text.split()]
            temp2.append(temp)
        float_posforces.append(temp2)

    out_dict = {}
    out_dict['displacements'] = float_displ
    out_dict['energies'] = energies
    out_dict['posforces'] = float_posforces

    return Dict(dict=out_dict)

def convert_frac(ratio):
    """ Converts ratio strings into float, e.g. 1.0/2.0 -> 0.5 """
    try:
        return float(ratio)
    except ValueError:
        num, denom = ratio.split('/')
        return float(num) / float(denom)






# write_results_to_file
from __future__ import absolute_import
def test_write_results_to_file_interface():
    """
    is the basic file writter wraper working as indended
    """
    from aiida_fleur.tools.io_routines import write_results_to_file
    from os.path import isfile, abspath
    from os import remove
    #import os
    from numpy import array
    

    # testing some defaults
    inputhead = 'head\n'
    data = array([[1,2],[3,4]])
    destination = './outputfiletest'
    write_results_to_file(inputhead, data, destination=destination)
    isfile_ = isfile(abspath('./outputfiletest'))
    test_file = open(destination, 'r')
    content = test_file.read()
    test_file.close()

    content_exp = 'head\n1.00000000  3.00000000\n2.00000000  4.00000000\n'
    remove(destination)

    assert isfile_
    assert content == content_exp


# write_xps_spectra_datafile
def test_write_xps_spectra_datafile_interface():
    """
    is the xps data file writter working, is the file ok?
    """
    from aiida_fleur.tools.io_routines import write_xps_spectra_datafile
    
    #TODO how to test this?
    # provide all sample inputs and check contents of outputfile
    pass
    #assert 1 == 2

def test_compress_fleuroutxml():
    """
    test the compress_fleuroutxml function, checks if right number of iterations is kept, or deleted.
    Further checks if new file is written and if eigenvalues are deleted.
    """
    
    from os.path import abspath, isfile
    from os import remove
    from lxml import etree
    from aiida_fleur.tools.xml_util import eval_xpath2
    from aiida_fleur.tools.io_routines import compress_fleuroutxml
    
    testfilepath = abspath('./files/outxml/BeTi_out.xml')
    dest_path = testfilepath.replace(".xml", "_test.xml")
    niter_file = 19 
    xpath_iter = '/fleurOutput/scfLoop/iteration'
    xpath_eig = '/fleurOutput/scfLoop/iteration/eigenvalues'
    
  
    def get_npath(filepath, xpath):
        """helper function to get amount of a certain xpath eval in file"""
        
        xpath_iter = '/fleurOutput/scfLoop/iteration'
        parser = etree.XMLParser(recover=False)
        tree = etree.parse(filepath, parser)

        return len(eval_xpath2(tree.getroot(), xpath))


    # test new file exists, and right number of iteration, eig del
    compress_fleuroutxml(testfilepath, dest_file_path=dest_path, iterations_to_keep=15)
    isfile_ = isfile(abspath(dest_path))
    niter1 = get_npath(dest_path, xpath_iter)
    neig = get_npath(dest_path, xpath_eig)    

    assert isfile                # check outfile
    assert niter1 == 15          # check if 15 iterations are kept
    assert neig==0               # check if eigenvalues del 
    
    compress_fleuroutxml(testfilepath, dest_file_path=dest_path, iterations_to_keep=-1)
    niter2 = get_npath(dest_path, xpath_iter)
    
    assert niter2 == 1           # check of only one iteration kept
     
    # test if more iteration given then in file should change nothing
    compress_fleuroutxml(testfilepath, dest_file_path=dest_path, iterations_to_keep=25)
    niter3 = get_npath(dest_path, xpath_iter)
    
    assert niter3 == niter_file  # check if no iteration deleted
    
    # cleanup
    remove(dest_path)




# test all routines used by the fleur parser

from __future__ import absolute_import
import os
import pytest

# parse_xmlout_file
def test_parse_xmlout_file():
    """
    tests if the routine that parsers the outputfile, produces the right output, no aiida datastructures,
    with the right content
    """
    from aiida_fleur.parsers.fleur import parse_xmlout_file
    import os

    filename = os.path.abspath('./files/outxml/BeTi_out.xml')

    simple_out, complex_out, parser_info_out, successful = parse_xmlout_file(filename)

    expected_simple_out_dict = {'bandgap': 0.0052350388,
                                 'bandgap_units': 'eV',
                                 'charge_den_xc_den_integral': -45.0947551412,
                                 'charge_density': 8.7984e-06,
                                 'creator_name': 'fleur 27',
                                 'creator_target_architecture': 'GEN',
                                 'creator_target_structure': ' ',
                                 'density_convergence_units': 'me/bohr^3',
                                 'energy': -23635.691764717936,
                                 'energy_core_electrons': -496.172547773,
                                 'energy_hartree': -868.5956587197,
                                 'energy_hartree_units': 'Htr',
                                 'energy_units': 'eV',
                                 'energy_valence_electrons': -7.1055909396,
                                 'fermi_energy': 0.3451127139,
                                 'fermi_energy_units': 'Htr',
                                 'force_largest': -0.0,
                                 'kmax': 4.5,
                                 'number_of_atom_types': 2,
                                 'number_of_atoms': 2,
                                 'number_of_iterations': 19,
                                 'number_of_iterations_total': 19,
                                 'number_of_kpoints': 56,
                                 'number_of_species': 1,
                                 'number_of_spin_components': 1,
                                 'number_of_symmetries': 48,
                                 'output_file_version': '0.27',
                                 'start_date': {'date': '2017/09/10', 'time': '07:58:10'},
                                 'end_date': {'date': '2017/09/10', 'time': '07:58:34'},
                                 'sum_of_eigenvalues': -503.2781387127,
                                 'title': 'Be-Ti, bulk compounds',
                                 'walltime': 24,
                                 'walltime_units': 'seconds',
                                 'warnings': {'debug': {}, 'error': {}, 'info': {}, 'warning': {}}}

    expected_parser_info_out = {'parser_info': 'AiiDA Fleur Parser v0.2beta',
                                'parser_warnings': [],
                                'unparsed': []}
    simple_out.pop('outputfile_path', None) # otherwise test will fail on different installations
    # also this should go away any way...

    assert successful == True
    assert expected_simple_out_dict == simple_out
    assert expected_parser_info_out == parser_info_out

# test special cases parser behavior
def test_parse_xmlout_file_broken_xmlout_file():
    """
    tests the behavior of the parse_xmlout_file routine in the case of an broken out.xml file.
    Here broken after serveral iteration, should parse all iteration except last.
    (which can happen in the case of some kill, or Non regular termination of FLEUR)
    """
    from aiida_fleur.parsers.fleur import parse_xmlout_file
    import os

    filename = os.path.abspath('./files/outxml/special/broken_BeTi_out.xml')

    simple_out, complex_out, parser_info_out, successful = parse_xmlout_file(filename)

    expected_parser_info_out = {
        'last_iteration_parsed': 15,
        'parser_info': 'AiiDA Fleur Parser v0.2beta',
        'parser_warnings': ['The out.xml file is broken I try to repair it.',
                           'Endtime was unparsed, inp.xml prob not complete, do not believe the walltime!'],
         'unparsed': []}

    assert successful == True
    assert 15 == parser_info_out['last_iteration_parsed']
    assert expected_parser_info_out['unparsed'] == parser_info_out['unparsed']
    assert expected_parser_info_out['parser_warnings'] == parser_info_out['parser_warnings']


def test_parse_xmlout_file_broken_first_xmlout_file():
    """
    tests the behavior of the parse_xmlout_file routine in the case of an broken out.xml file.
    Here broken in first iteration, should parse nothing.
    (which can happen in the case of some kill, or Non regular termination of FLEUR)
    """
    from aiida_fleur.parsers.fleur import parse_xmlout_file
    import os

    filename = os.path.abspath('./files/outxml/special/broken_first_BeTi_out.xml')

    simple_out, complex_out, parser_info_out, successful = parse_xmlout_file(filename)

    expected_parser_info_out = {
         'last_iteration_parsed': 1,
         'parser_info': 'AiiDA Fleur Parser v0.2beta',
         'parser_warnings': ['The out.xml file is broken I try to repair it.',
                             'Can not get attributename: "units" from node "[]", because node is not an element of etree.',
                             'Can not get attributename: "value" from node "[]", because node is not an element of etree.',
                             'Could not convert: "None" to float, TypeError',
                             'Could not convert: "None" to float, TypeError',
                             'Could not convert: "None" to float, TypeError',
                             'Can not get attributename: "value" from node "[]", because node is not an element of etree.',
                             'Could not convert: "None" to float, TypeError',
                             'Can not get attributename: "value" from node "[]", because node is not an element of etree.',
                             'Could not convert: "None" to float, TypeError',
                             'Can not get attributename: "value" from node "[]", because node is not an element of etree.',
                             'Could not convert: "None" to float, TypeError',
                             'Can not get attributename: "value" from node "[]", because node is not an element of etree.',
                             'Could not convert: "None" to float, TypeError',
                             'Can not get attributename: "units" from node "[]", because node is not an element of etree.',
                             'Can not get attributename: "value" from node "[]", because node is not an element of etree.',
                             'Could not convert: "None" to float, TypeError',
                             'Can not get attributename: "value" from node "[]", because node is not an element of etree.',
                             'Could not convert: "None" to float, TypeError',
                             'Can not get attributename: "units" from node "[]", because node is not an element of etree.',
                             'Can not get attributename: "units" from node "[]", because node is not an element of etree.',
                             'Can not get attributename: "distance" from node "[]", because node is not an element of etree.',
                             'Could not convert: "None" to float, TypeError',
                             'Endtime was unparsed, inp.xml prob not complete, do not believe the walltime!'],
         'unparsed': [{'energy_hartree': None, 'iteration': '    1'},
                      {'energy': None, 'iteration': '    1'},
                      {'iteration': '    1', 'sum_of_eigenvalues': None},
                      {'energy_core_electrons': None, 'iteration': '    1'},
                      {'energy_valence_electrons': None, 'iteration': '    1'},
                      {'charge_den_xc_den_integral': None, 'iteration': '    1'},
                      {'bandgap': None, 'iteration': '    1'},
                      {'fermi_energy': None, 'iteration': '    1'},
                      {'charge_density': None, 'iteration': '    1'}]}

    assert successful == True
    assert 1 == parser_info_out['last_iteration_parsed']
    assert expected_parser_info_out['unparsed'] == parser_info_out['unparsed']
    assert expected_parser_info_out['parser_warnings'] == parser_info_out['parser_warnings']


def test_parse_xmlout_file_fortran_garbage_in_xmlout_file():
    """
    tests the behavior of the parse_xmlout_file routine in the case of an individual 'garbage' in the out.xml file.
    Fortran NANs and INFs will be parsed fine, ** will not be parsed.
    (which can happen in the case of some kill, or Non regular termination of FLEUR)
    """
    from aiida_fleur.parsers.fleur import parse_xmlout_file
    import os

    filename = os.path.abspath('./files/outxml/special/Fortran_garbage_BeTi_out.xml')

    simple_out, complex_out, parser_info_out, successful = parse_xmlout_file(filename)

    exp_partial_simple_out_dict = {
        'bandgap_units': 'eV',
        'energy': float('Inf'),
        'energy_hartree':  float('Inf'),
        'fermi_energy': float('NaN'),
        'warnings': {'debug': {}, 'error': {}, 'info': {}, 'warning': {}}}

    expected_parser_info_out = {
         'parser_info': 'AiiDA Fleur Parser v0.2beta',
         'parser_warnings': ['Could not convert: "**" to float, ValueError',
                             'Could not convert: "        !#@)!(U$*(Y" to float, ValueError'],
         'unparsed': [{'bandgap': '**', 'iteration': '   19'},
                      {'charge_density': '        !#@)!(U$*(Y', 'iteration': '   19'}]}

    #TODO maybe in the case on unpared, things should be initialized, here they are missing...
    def isNaN(num):
        return num != num

    assert successful == True
    assert exp_partial_simple_out_dict['energy'] == simple_out['energy']
    assert exp_partial_simple_out_dict['energy_hartree'] == simple_out['energy_hartree']
    assert isNaN(exp_partial_simple_out_dict['fermi_energy']) == isNaN(simple_out['fermi_energy'])
    assert 'bandgap' not in list(simple_out.keys())

    assert expected_parser_info_out['unparsed'] == parser_info_out['unparsed']
    assert expected_parser_info_out['parser_warnings'] == parser_info_out['parser_warnings']


def test_parse_xmlout_file_empty_file():
    """
    tests the behavior of the parse_xmlout_file routine in the case of an empty file
    """
    from aiida_fleur.parsers.fleur import parse_xmlout_file
    import os

    filename = os.path.abspath('./files/outxml/special/empty_out.xml')

    simple_out, complex_out, parser_info_out, successful = parse_xmlout_file(filename)

    expected_parser_info_out = {
         'parser_info': 'AiiDA Fleur Parser v0.2beta',
         'parser_warnings': ['The out.xml file is broken I try to repair it.',
                             'Skipping the parsing of the xml file. Repairing was not possible.'],
         'unparsed': []}

    assert successful == False
    assert expected_parser_info_out == parser_info_out


# test parser success for all out files in folder
file_path1 = '../files/outxml/all_test/'
outxmlfilefolder = os.path.dirname(os.path.abspath(__file__))
outxmlfilefolder_valid =  os.path.abspath(os.path.join(outxmlfilefolder, file_path1))

outxmlfilelist = []
for subdir, dirs, files in os.walk(outxmlfilefolder_valid):
    for file in files:
        if file.endswith('.xml'):
            outxmlfilelist.append(os.path.join(subdir, file))


@pytest.mark.parametrize("xmloutfile", outxmlfilelist)
def test_fleurparse_all_xmlout_file(xmloutfile):
    """
    tests if the routine that parsers the outputfile, succeeds for all out files
    """
    from aiida_fleur.parsers.fleur import parse_xmlout_file

    simple_out, complex_out, parser_info_out, successful = parse_xmlout_file(xmloutfile)

    assert successful == True


# parse_dos_file, test for different dos files with spin and without
def parse_dos_file():
    """
    test for the fleur dos file parser. test if right output, datastructures are produced without error
    """
    from aiida_fleur.parsers.fleur import parse_dos_file
    pass

    # test if array data is prodcued without error


# parse_bands_file
def test_parse_bands_file():
    """
    test for band file parse routine.
    """

    from aiida_fleur.parsers.fleur import parse_bands_file
    pass

    # test if a bandsdata object is produced




# test the full parser itself. on all kinds of different output files.

# test if the right aiida datastructures are produced for different output
# also check if errors are working...
# if an empty and broken file works, broken before and after first iteration





# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################

# Here we test if the interfaces of the workflows are still the same
from __future__ import absolute_import
import pytest


@pytest.mark.usefixtures("aiida_profile", "clear_database")
class TestFleurWorkchainInterfaces:
    """
    Test all aiida-fleur workflow interfaces
    """
    # TODO
    # prepare some nodes:
    # structure, option, fleurinp, wfparameters
    # add to builder and see if he takes it
    # ggf if possible run initial step only, that the input is checked...
    # In general the interfaces should be fixed and not changed. this is what
    # these tests are for, to test be aware of interface breaks

    def test_fleur_scf_wc_init(self):
        """
        Test the interface of the scf workchain
        """
        from aiida_fleur.workflows.scf import FleurScfWorkChain

        builder = FleurScfWorkChain.get_builder()

    def test_fleur_eos_wc_init(self):
        """
        Test the interface of the eos workchain
        """
        from aiida_fleur.workflows.eos import FleurEosWorkChain

        builder = FleurEosWorkChain.get_builder()

    def test_fleur_dos_wc_init(self):
        """
        Test the interface of the dos workchain
        """
        from aiida_fleur.workflows.dos import fleur_dos_wc

        builder = fleur_dos_wc.get_builder()

    def test_fleur_band_wc_init(self):
        """
        Test the interface of the band workchain
        """
        from aiida_fleur.workflows.scf import FleurScfWorkChain

        builder = FleurScfWorkChain.get_builder()

    # def test_fleur_band2_wc_init(self):
    #    """
    #    Test the interface of the band2 workchain
    #    """
    #    from aiida_fleur.workflows.band2 import fleur_band2_wc
    #
    #    builder = fleur_band2_wc.get_builder()

    def test_fleur_corehole_wc_init(self):
        """
        Test the interface of the corehole workchain
        """
        from aiida_fleur.workflows.corehole import fleur_corehole_wc

        builder = fleur_corehole_wc.get_builder()

    def test_fleur_initial_cls_wc_init(self):
        """
        Test the interface of the scf workchain
        """
        from aiida_fleur.workflows.initial_cls import fleur_initial_cls_wc

        builder = fleur_initial_cls_wc.get_builder()

    def test_fleur_delta_wc_init(self):
        """
        Test the interface of the delta workchain
        """
        from aiida_fleur.workflows.delta import fleur_delta_wc

        builder = fleur_delta_wc.get_builder()

    def test_fleur_relax_wc_init(self):
        """
        Test the interface of the relax workchain
        """
        from aiida_fleur.workflows.relax import FleurRelaxWorkChain

        builder = FleurRelaxWorkChain.get_builder()

    def test_fleur_optimize_para_wc_init(self):
        """
        Test the interface of the optimize_para_ workchain
        """
        from aiida_fleur.workflows.optimize_para import fleur_optimize_parameters_wc

        builder = fleur_optimize_parameters_wc.get_builder()

    def test_fleur_mae_wc_init(self):
        """
        Test the interface of the dmi workchain
        """
        from aiida_fleur.workflows.mae import FleurMaeWorkChain

        builder = FleurMaeWorkChain.get_builder()

    def test_fleur_mae_conv_wc_init(self):
        """
        Test the interface of the dmi workchain
        """
        from aiida_fleur.workflows.mae_conv import FleurMaeConvWorkChain

        builder = FleurMaeConvWorkChain.get_builder()

    def test_fleur_ssdisp_wc_init(self):
        """
        Test the interface of the dmi workchain
        """
        from aiida_fleur.workflows.ssdisp import FleurSSDispWorkChain

        builder = FleurSSDispWorkChain.get_builder()

    def test_fleur_ssdisp_conv_wc_init(self):
        """
        Test the interface of the dmi workchain
        """
        from aiida_fleur.workflows.ssdisp_conv import FleurSSDispConvWorkChain

        builder = FleurSSDispConvWorkChain.get_builder()

    def test_fleur_dmi_wc_init(self):
        """
        Test the interface of the dmi workchain
        """
        from aiida_fleur.workflows.dmi import FleurDMIWorkChain

        builder = FleurDMIWorkChain.get_builder()

    def test_fleur_base_wc_init(self):
        """
        Test the interface of the dmi workchain
        """
        from aiida_fleur.workflows.base_fleur import FleurBaseWorkChain

        builder = FleurBaseWorkChain.get_builder()

    def test_fleur_base_relax_wc_init(self):
        """
        Test the interface of the dmi workchain
        """
        from aiida_fleur.workflows.base_relax import FleurBaseRelaxWorkChain

        builder = FleurBaseRelaxWorkChain.get_builder()

    def test_fleur_create_magnetic_wc_init(self):
        """
        Test the interface of the dmi workchain
        """
        from aiida_fleur.workflows.create_magnetic_film import FleurCreateMagneticWorkChain

        builder = FleurCreateMagneticWorkChain.get_builder()
